"""Top-level class for all data concepts objects and collections thereof."""
from abc import abstractmethod, ABC
import re
from typing import TypeVar, Type, List, Union, Optional, Iterator, Iterable
from uuid import UUID, uuid4

from gemd.entity.dict_serializable import DictSerializable, DictSerializableMeta
from gemd.entity.base_entity import BaseEntity
from gemd.entity.link_by_uid import LinkByUID
from gemd.json import GEMDJson
from gemd.util import recursive_foreach, set_uuids

from citrine._rest.collection import Collection
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._serialization.properties import String, Mapping, Object
from citrine._serialization.properties import Optional as PropertyOptional
from citrine._serialization.properties import List as PropertyList
from citrine._serialization.properties import UUID as PropertyUUID
from citrine._serialization.serializable import Serializable
from citrine._session import Session
from citrine._utils.functions import scrub_none, replace_objects_with_links, \
    format_escaped_url
from citrine.exceptions import BadRequest
from citrine.resources.audit_info import AuditInfo
from citrine.jobs.job import _poll_for_job_completion
from citrine.resources.response import Response

CITRINE_SCOPE = 'id'
CITRINE_TAG_PREFIX = 'citr_auto'


class DataConceptsMeta(DictSerializableMeta):
    """Data Concepts metaclass for handling serialization."""

    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        resolved = next((b.typ for b in cls.__bases__ if getattr(b, "typ", None) is not None),
                        None)
        if resolved is not None:
            cls._typ_stash = resolved
        cls.typ = String("type")


class DataConcepts(
    PolymorphicSerializable['DataConcepts'],
    BaseEntity,
    metaclass=DataConceptsMeta
):
    """
    An abstract data concepts object.

    DataConcepts must be extended along with `Resource`.

    Parameters
    ----------
    typ: str
        A string denoting what type of DataConcepts class a particular instantiation is.

    """

    """Properties inherited from GEMD Base Entitiy."""
    uids = PropertyOptional(Mapping(String('scope'), String('id')), 'uids', override=True)
    tags = PropertyOptional(PropertyList(String()), 'tags', override=True)

    _type_key = "type"
    """str: key used to determine type of serialized object."""

    collection_dict = dict()
    """
    Dict[str, class]: dictionary from the type key to the associated collection \
     for every class that extends DataConcepts.

    Only populated if the :func:`get_collection_type` method is invoked.
    """

    """
    Fields that are added to the gemd data objects when they are used in this client

    * audit_info contains who/when information about the resource on the citrine platform
    * dataset is the unique Citrine id of the dataset that owns this resource
    """
    _audit_info = PropertyOptional(Object(AuditInfo), "audit_info", serializable=False)
    _dataset = PropertyOptional(PropertyUUID, "dataset", serializable=False)

    def __init__(self):
        self.typ = self._typ_stash

    @property
    def audit_info(self) -> Optional[AuditInfo]:
        """Get the audit info object."""
        return self._audit_info

    @property
    def uid(self) -> Optional[UUID]:
        """Get the Citrine Identifier (scope = "id"), or None if not registered."""
        return self.uids.get(CITRINE_SCOPE)

    @property
    def dataset(self) -> Optional[UUID]:
        """Get the dataset of this object, if it was returned by the backend."""
        return self._dataset

    @classmethod
    def get_type(cls, data) -> Type[Serializable]:
        """
        Determine the class of a serialized object.

        The data dictionary must have a 'type' key whose value corresponds to the response key
        of one of the classes that extends :class:`DataConcepts`.

        Parameters
        ----------
        data: dict
            A dictionary corresponding to a serialized data concepts object of unknown type.
            This method will also work if data is a deserialized GEMD object.

        Returns
        -------
        class
            The class corresponding to data.

        """
        if isinstance(data, DictSerializable):
            data = data.as_dict()
        return DictSerializable.class_mapping[data['type']]

    @classmethod
    def get_collection_type(cls, data) -> "Type[DataConceptsCollection]":
        """
        Determine the associated collection type for a serialized data object.

        The data dictionary must have a 'type' key whose value corresponds to the individual key
        of one of the collections that extends :class:`DataConceptsCollection`.

        Parameters
        ----------
        data: dict
            A dictionary corresponding to a serialized data concepts object of unknown type.
            This method will also work if data is a deserialized GEMD object.

        Returns
        -------
        collection
            The collection type corresponding to data.

        """
        if len(DataConcepts.collection_dict) == 0:
            # This branch is intended to populate collection dict on first call
            DataConcepts._make_collection_dict()
        if isinstance(data, DictSerializable):
            data = data.as_dict()
        return DataConcepts.collection_dict[data['type']]

    @staticmethod
    def _make_collection_dict():
        """Construct a dictionary from each type key to the associated collection."""
        from citrine.resources.condition_template import ConditionTemplateCollection
        from citrine.resources.parameter_template import ParameterTemplateCollection
        from citrine.resources.property_template import PropertyTemplateCollection
        from citrine.resources.material_template import MaterialTemplateCollection
        from citrine.resources.measurement_template import MeasurementTemplateCollection
        from citrine.resources.process_template import ProcessTemplateCollection
        from citrine.resources.ingredient_spec import IngredientSpecCollection
        from citrine.resources.material_spec import MaterialSpecCollection
        from citrine.resources.measurement_spec import MeasurementSpecCollection
        from citrine.resources.process_spec import ProcessSpecCollection
        from citrine.resources.ingredient_run import IngredientRunCollection
        from citrine.resources.material_run import MaterialRunCollection
        from citrine.resources.measurement_run import MeasurementRunCollection
        from citrine.resources.process_run import ProcessRunCollection
        _collection_list = [
            ConditionTemplateCollection, ParameterTemplateCollection, PropertyTemplateCollection,
            MaterialTemplateCollection, MeasurementTemplateCollection, ProcessTemplateCollection,
            IngredientSpecCollection, MaterialSpecCollection, MeasurementSpecCollection,
            ProcessSpecCollection, IngredientRunCollection, MaterialRunCollection,
            MeasurementRunCollection, ProcessRunCollection
        ]
        for collection in _collection_list:
            DataConcepts.collection_dict[collection._individual_key] = collection


def _make_link_by_uid(gemd_object_rep: Union[str, UUID, BaseEntity, LinkByUID]) -> LinkByUID:
    if isinstance(gemd_object_rep, BaseEntity):
        return gemd_object_rep.to_link(CITRINE_SCOPE, allow_fallback=True)
    elif isinstance(gemd_object_rep, LinkByUID):
        return gemd_object_rep
    elif isinstance(gemd_object_rep, (str, UUID)):
        uid = str(gemd_object_rep)
        scope = CITRINE_SCOPE
        return LinkByUID(scope, uid)
    else:
        raise TypeError("Link can only be created from a GEMD object, LinkByUID, str, or UUID."
                        "Instead got {}.".format(gemd_object_rep))


ResourceType = TypeVar('ResourceType', bound='DataConcepts')


class DataConceptsCollection(Collection[ResourceType], ABC):
    """
    A collection of one kind of data concepts object.

    Parameters
    ----------
    project_id: UUID
        The uid of the project that this collection belongs to.
    dataset_id: UUID
        The uid of the dataset that this collection belongs to. If None then the collection
        ranges over all datasets in the project. Note that this is only allowed for certain
        actions. For example, you can use :func:`list_by_tag` to search over all datasets,
        but when using :func:`register` to upload or update an object, a dataset must be specified.
    session: Session
        The Citrine session used to connect to the database.

    """

    def __init__(self, project_id: UUID, dataset_id: UUID, session: Session):
        self.project_id = project_id
        self.dataset_id = dataset_id
        self.session = session

    @classmethod
    @abstractmethod
    def get_type(cls) -> Type[Serializable]:
        """Return the resource type in the collection."""

    def build(self, data: dict) -> ResourceType:
        """
        Build an object of type ResourceType from a serialized dictionary.

        This is an internal method, and should not be called directly by users.

        Parameters
        ----------
        data: dict
            A serialized data model object.

        Returns
        -------
        ResourceType
            A data model object built from the dictionary.

        """
        return self.get_type().build(data)

    def list(self, *,
             per_page: Optional[int] = 100,
             forward: bool = True) -> Iterator[ResourceType]:
        """
        Get all visible elements of the collection.

        The order of results should not be relied upon, but for now they are sorted by
        dataset, object type, and creation time (in that order of priority).

        Parameters
        ---------
        per_page: int, optional
            Max number of results to return per page. It is very unlikely that
            setting this parameter to something other than the default is useful.
            It exists for rare situations where the client is bandwidth constrained
            or experiencing latency from large payload sizes.
        forward: bool
            Set to False to reverse the order of results (i.e., return in descending order)

        Returns
        -------
        Iterator[ResourceType]
            Every object in this collection.

        """
        params = {}
        if self.dataset_id is not None:
            params['dataset_id'] = str(self.dataset_id)
        raw_objects = self.session.cursor_paged_resource(
            self.session.get_resource,
            self._get_path(ignore_dataset=True),
            forward=forward,
            per_page=per_page,
            params=params)
        return (self.build(raw) for raw in raw_objects)

    def register(self, model: ResourceType, *, dry_run=False):
        """
        Create a new element of the collection or update an existing element.

        If the input model has an ID that corresponds to an existing object in the
        database, then that object will be updated. Otherwise a new object will be created.

        Only the top-level object in `model` itself is written to the database with this
        method. References to other objects are persisted as links, and the object returned
        by this method has all instances of data objects replaced by instances of LinkByUid.
        Registering an object which references other objects does NOT implicitly register
        those other objects. Rather, those other objects' values are ignored, and the
        pre-existence of objects with their IDs is asserted before attempting to write
        `model`.

        Parameters
        ----------
        model: ResourceType
            The DataConcepts object.
        dry_run: bool
            Whether to actually register the item or run a dry run of the register operation.
            Dry run is intended to be used for validation. Default: false

        Returns
        -------
        ResourceType
            A copy of the registered object as it now exists in the database.

        """
        if self.dataset_id is None:
            raise RuntimeError("Must specify a dataset in order to register a data model object.")
        path = self._get_path()
        params = {'dry_run': dry_run}

        temp_scope = str(uuid4())
        scope = temp_scope if dry_run else CITRINE_SCOPE
        set_uuids(model, scope=scope)
        dumped_data = replace_objects_with_links(scrub_none(model.dump()))

        data = self.session.post_resource(path, dumped_data, params=params)
        registered = self.build(data)

        recursive_foreach(model, lambda x: x.uids.pop(temp_scope, None))  # Strip temp uids
        if not dry_run:
            # Platform may add a CITRINE_SCOPE uid and citr_auto tags; update locals
            model.uids.update({k: v for k, v in registered.uids.items()})
            if registered.tags is not None:
                if model.tags is None:  # This is somehow hit by nextgen-devkit tests
                    model.tags = list()  # pragma: no cover
                model.tags.extend([tag for tag in registered.tags
                                   if re.match(f"^{CITRINE_TAG_PREFIX}::", tag)])
        else:
            # Remove of the tags/uids the platform spuriously added
            # this might leave objects with just the temp ids, which we want to strip later
            if CITRINE_SCOPE not in model.uids:
                registered.uids.pop(CITRINE_SCOPE, None)
            if registered.tags is not None:
                todo = [tag for tag in registered.tags
                        if re.match(f"^{CITRINE_TAG_PREFIX}::", tag)]
                for tag in todo:  # Covering this block would require dark art
                    if tag not in model.tags:
                        registered.tags.remove(tag)
        return registered

    def register_all(self,
                     models: Iterable[ResourceType],
                     *,
                     dry_run: bool = False,
                     status_bar: bool = False,
                     include_nested: bool = False) -> List[ResourceType]:
        """
        Register multiple GEMD objects to each of their appropriate collections.

        Does so in an order that is guaranteed to store all linked items before the item that
        references them.

        If the GEMD objects have no UIDs, Citrine IDs will be assigned to them prior to passing
        them on to the server.  This is required as otherwise there is no way to determine how
        objects are related to each other.  When the registered objects are returned from the
        server, the input GEMD objects will be updated with whichever uids & _citr_auto:: tags are
        on the returned objects.  This means GEMD objects that already exist on the server will
        be updated with all their on-platform uids and tags.

        This method has the same behavior as `register`, except that no models will be
        written if any one of them is invalid.  Using this method should yield significant
        improvements to write speed over separate calls to `register`.

        Parameters
        ----------
        models: Iterable[DataConcepts]
            The data model objects to register. Can be different types.

        dry_run: bool
            Whether to actually register the objects or run a dry run of the register operation.
            Dry run is intended to be used for validation. Default: false

        status_bar: bool
            Whether to display a status bar using the tqdm module to track progress in
            registration. Requires installing the optional tqdm module. Default: false

        include_nested: bool
            Whether to just register the objects passed in the list, or include nested objects
            (e.g., obj.process, obj.spec.template, ...).  Default: false

        Returns
        -------
        List[DataConcepts]
            Each object model as it now exists in the database.

        """
        from citrine.resources.gemd_resource import GEMDResourceCollection
        gemd_collection = GEMDResourceCollection(self.project_id, self.dataset_id, self.session)
        return gemd_collection.register_all(
            models,
            dry_run=dry_run,
            status_bar=status_bar,
            include_nested=include_nested
        )

    def update(self, model: ResourceType) -> ResourceType:
        """
        Update a data object model.

        Update a particular element of the collection,
        first attempting a simple update using register and falling back to async_update
        if the changes require it (e.g., updating template bounds).

        model: ResourceType
            The DataConcepts object.
        """
        try:
            return self.register(model, dry_run=False)
        except BadRequest:
            # If register() cannot be used because an asynchronous check is required
            return self.async_update(model, dry_run=False,
                                     wait_for_response=True, return_model=True)

    def async_update(self, model: ResourceType, *,
                     dry_run: bool = False,
                     wait_for_response: bool = True,
                     timeout: float = 2 * 60,
                     polling_delay: float = 1.0,
                     return_model: bool = False) -> Optional[Union[UUID, ResourceType]]:
        """
        Update a particular element of the collection with data validation.

        Update a particular element of the collection, doing a deeper check to ensure that
        the dependent data objects are still with the (potentially) changed constraints
        of this change. This will allow you to make bounds and allowed named/labels changes
        to templates.

        Parameters
        ----------
        model: ResourceType
            The DataConcepts object.
        dry_run: bool
            Whether to actually update the item or run a dry run of the update operation.
            Dry run is intended to be used for validation. Default: false
        wait_for_response: bool
            Whether to poll for the eventual response. This changes the return type (see
            below).
        timeout: float
            How long to poll for the result before giving up. This is expressed in
            (fractional) seconds.
        polling_delay: float
            How long to delay between each polling retry attempt.
        return_model: bool
            Whether or not to return an updated version of the resource
            If wait_for_response is False, then this argument has no effect

        Returns
        -------
        Optional[UUID]
            If wait_for_response if True, then this call will poll the backend, waiting
            for the eventual job result. In the case of successful validation/update,
            a return value of None is provided unless return_model is True, in which case
            the updated resource is fetched and returned. In the case of a failure
            validating or processing the update, an exception (JobFailureError) is raised
            and an error message is logged with the underlying reason of the failure.

            If wait_for_response if False, A job ID (of type UUID) is returned that one
            can use to poll for the job completion and result with the
            :func:`~citrine.resources.DataConceptsCollection.poll_async_update_job`
            method.

        """
        temp_scope = str(uuid4())
        GEMDJson(scope=temp_scope).dumps(model)  # This apparent no-op populates uids
        dumped_data = replace_objects_with_links(scrub_none(model.dump()))
        recursive_foreach(model, lambda x: x.uids.pop(temp_scope, None))  # Strip temp uids

        scope = CITRINE_SCOPE
        id = dumped_data['uids'][scope]
        if self.dataset_id is None:
            raise RuntimeError("Must specify a dataset in order to update "
                               "a data model object with data validation.")

        url = self._get_path(action=[scope, id, "async"])
        response_json = self.session.put_resource(url, dumped_data, params={'dry_run': dry_run})

        job_id = response_json["job_id"]

        if wait_for_response:
            self.poll_async_update_job(job_id, timeout=timeout,
                                       polling_delay=polling_delay)

            # That worked, return nothing or return the object
            if return_model:
                return self.get(LinkByUID(scope=scope, id=id))
            else:
                return None
        else:
            # TODO: use JobSubmissionResponse here instead
            return job_id

    def poll_async_update_job(self, job_id: UUID, *, timeout: float = 2 * 60,
                              polling_delay: float = 1.0) -> None:
        """
        Poll for the result of the async_update call.

        This call will poll the backend given the Job ID that came from a call to
        :func:`~citrine.resources.DataConceptsCollection.async_update`,
        waiting for the eventual job result. In the case of successful validation/update,
        a return value of None is provided which indicates success. In the case of
        a failure validating or processing the update, an exception (JobFailureError)
        is raised and an error message is logged with the underlying reason of the
        failure.

        Parameters
        ----------
        job_id: UUID
           The job ID for the asynchronous update job we wish to poll.
        timeout:
            How long to poll for the result before giving up. This is expressed in
            (fractional) seconds.
        polling_delay:
            How long to delay between each polling retry attempt.

        Returns
        -------
        None
           This method will raise an appropriate exception if the job failed, else
           it will return None to indicate the job was successful.

        """
        # Poll for job completion - this will raise an error if the job failed
        _poll_for_job_completion(self.session, self.project_id, job_id, timeout=timeout,
                                 polling_delay=polling_delay)

        # That worked, nothing returned in this case
        return None

    def get(self, uid: Union[UUID, str, LinkByUID, BaseEntity]) -> ResourceType:
        """
        Get an element of the collection by its id.

        Parameters
        ----------
        uid: Union[UUID, str, LinkByUID, BaseEntity]
            A representation of the object (Citrine id, LinkByUID, or the object itself)

        Returns
        -------
        ResourceType
            An object with specified scope and uid

        """
        link = _make_link_by_uid(uid)
        path = self._get_path(ignore_dataset=self.dataset_id is None, action=[link.scope, link.id])
        data = self.session.get_resource(path)
        return self.build(data)

    def list_by_name(self, name: str, *, exact: bool = False,
                     forward: bool = True, per_page: int = 100) -> Iterator[ResourceType]:
        """
        Get all objects with specified name in this dataset.

        Parameters
        ----------
        name: str
            case-insensitive object name prefix to search.
        exact: bool
            Set to True to change prefix search to exact search (but still case-insensitive).
            Default is False.
        forward: bool
            Set to False to reverse the order of results (i.e., return in descending order).
        per_page: int
            Controls the number of results fetched with each http request to the backend.
            Typically, this is set to a sensible default and should not be modified. Consider
            modifying this value only if you find this method is unacceptably latent.

        Returns
        -------
        Iterator[ResourceType]
            List of every object in this collection whose `name` matches the search term.

        """
        if self.dataset_id is None:
            raise RuntimeError("Must specify a dataset to filter by name.")
        params = {'dataset_id': str(self.dataset_id), 'name': name, 'exact': exact}
        raw_objects = self.session.cursor_paged_resource(
            self.session.get_resource,
            # "Ignoring" dataset because it is in the query params (and required)
            self._get_path(ignore_dataset=True, action="filter-by-name"),
            forward=forward,
            per_page=per_page,
            params=params)
        return (self.build(raw) for raw in raw_objects)

    def list_by_tag(self, tag: str, *, per_page: int = 100) -> Iterator[ResourceType]:
        """
        Get all objects bearing a tag prefixed with `tag` in the collection.

        The order of results is largely not meaningful. Results from the same dataset will be
        grouped together but no other meaningful ordering can be relied upon. Duplication in
        the result set may (but needn't) occur when one object has multiple tags matching the
        search tag. For this reason, it is inadvisable to put 2 tags with the same prefix
        (e.g., 'foo::bar' and 'foo::baz') in the same object when it can be avoided.

        Parameters
        ----------
        tag: str
            The prefix with which to search. Must fully match up to the first delimiter (ex.
            'foo' and 'foo::b' both match 'foo::bar' but 'fo' is insufficient.
        per_page: int
            Controls the number of results fetched with each http request to the backend.
            Typically, this is set to a sensible default and should not be modified. Consider
            modifying this value only if you find this method is unacceptably latent.

        Returns
        -------
        Iterator[ResourceType]
            Every object in this collection.

        """
        params = {'tags': [tag]}
        if self.dataset_id is not None:
            params['dataset_id'] = str(self.dataset_id)
        raw_objects = self.session.cursor_paged_resource(
            self.session.get_resource,
            self._get_path(ignore_dataset=True),
            per_page=per_page,
            params=params)
        return (self.build(raw) for raw in raw_objects)

    def delete(self, uid: Union[UUID, str, LinkByUID, BaseEntity], *, dry_run: bool = False):
        """
        Delete an element of the collection by its id.

        Parameters
        ----------
        uid: Union[UUID, str, LinkByUID, BaseEntity]
            A representation of the object (Citrine id, LinkByUID, or the object itself)
        dry_run: bool
            Whether to actually delete the item or run a dry run of the delete operation.
            Dry run is intended to be used for validation. Default: false

        """
        link = _make_link_by_uid(uid)
        path = self._get_path(action=[link.scope, link.id])
        params = {'dry_run': dry_run}
        self.session.delete_resource(path, params=params)
        return Response(status_code=200)  # delete succeeded

    def _get_relation(self, relation: str, uid: Union[UUID, str, LinkByUID, BaseEntity],
                      forward: bool = True, per_page: int = 100) -> Iterator[ResourceType]:
        """
        Generic method for searching this collection by relation to another object.

        Parameters
        ----------
        relation
            Reflects the type of the object with the provided uid and scope, e.g.,
            'process-templates' if searching for process specs by process template.
        uid
            A representation of the object upon which this search is based, e.g., a
            Citrine ID of a process template whose process spec usages are being located.
        forward
            Whether to pages results in ascending order. Typically this is an
            unnecessary parameter.
        per_page
            The number of results to retrieve in each request to the backend. Typically
            this is an unnecessary parameter.
        Returns
        -------
        Iterator[ResourceType]
            Objects in this collection which are somehow related to the object with
            provided uid and scope.

        """
        params = {}
        if self.dataset_id is not None:
            params['dataset_id'] = str(self.dataset_id)
        link = _make_link_by_uid(uid)
        raw_objects = self.session.cursor_paged_resource(
            self.session.get_resource,
            format_escaped_url('projects/{}/{}/{}/{}/{}',
                               self.project_id,
                               relation,
                               link.scope,
                               link.id,
                               self._collection_key.replace('_', '-')
                               ),
            forward=forward,
            per_page=per_page,
            params=params,
            version='v1')
        return (self.build(raw) for raw in raw_objects)



from functools import partial
from typing import Iterable
from uuid import UUID

from citrine._rest.resource import Resource
from citrine._utils.functions import format_escaped_url
from citrine.informatics.generative_design import GenerativeDesignResult
from citrine.informatics.executions.execution import Execution


class GenerativeDesignExecution(Resource["GenerativeDesignExecution"], Execution):
    """The execution of a Generative Design task.

    Possible statuses are INPROGRESS, SUCCEEDED, and FAILED.
    Design executions also have a ``status_description`` field with more information.

    """

    def _path(self):
        return format_escaped_url(
            '/projects/{project_id}/generative-design/executions/',
            project_id=self.project_id,
        )

    @classmethod
    def _build_results(
        cls, subset_collection: Iterable[dict]
    ) -> Iterable[GenerativeDesignResult]:
        for generation_result in subset_collection:
            yield GenerativeDesignResult.build(generation_result)

    def results(self, *, per_page: int = 100) -> Iterable[GenerativeDesignResult]:
        """Fetch the Generative Design Results for the particular execution, paginated."""
        path = self._path() + f'{self.uid}/results'
        fetcher = partial(self._fetch_page, path=path, fetch_func=self._session.get_resource)
        return self._paginator.paginate(page_fetcher=fetcher,
                                        collection_builder=self._build_results,
                                        per_page=per_page)

    def result(
        self,
        *,
        result_id: UUID,
    ) -> GenerativeDesignResult:
        """Fetch a Generative Design Result for the particular UID."""
        path = self._path() + f'{self.uid}/results/{result_id}'
        data = self._session.get_resource(path, version=self._api_version)
        result = GenerativeDesignResult.build(data)
        return result



from typing import List, Mapping, Any

from citrine._rest.engine_resource import ModuleEngineResource
from citrine._serialization import properties
from citrine.informatics.descriptors import Descriptor
from citrine.informatics.design_spaces.design_space import DesignSpace

__all__ = ['EnumeratedDesignSpace']


class EnumeratedDesignSpace(ModuleEngineResource['EnumeratedDesignSpace'], DesignSpace):
    """An explicit enumeration of candidate materials to score.

    Enumerated design spaces are intended to capture small spaces with fewer than
    1000 values.  For larger spaces, use the DataSourceDesignSpace.

    Parameters
    ----------
    name:str
        the name of the design space
    description:str
        the description of the design space
    descriptors: list[Descriptor]
        the list of descriptors included in the candidates of the design space
    data: list[dict]
        list of dicts of the shape `{<descriptor_key>: <descriptor_value>}`
        where each dict corresponds to a candidate in the design space

    """

    descriptors = properties.List(properties.Object(Descriptor), 'data.instance.descriptors')
    data = properties.List(properties.Mapping(properties.String, properties.Raw),
                           'data.instance.data')

    typ = properties.String('data.instance.type', default='EnumeratedDesignSpace',
                            deserializable=False)

    def __init__(self,
                 name: str,
                 *,
                 description: str,
                 descriptors: List[Descriptor],
                 data: List[Mapping[str, Any]]):
        self.name: str = name
        self.description: str = description
        self.descriptors: List[Descriptor] = descriptors
        self.data: List[Mapping[str, Any]] = data

    def __str__(self):
        return '<EnumeratedDesignSpace {!r}>'.format(self.name)



from functools import lru_cache

from citrine._serialization import properties
from citrine._utils.functions import format_escaped_url
from citrine.informatics.predictor_evaluation_result import PredictorEvaluationResult
from citrine.informatics.executions.execution import Execution
from citrine._rest.resource import Resource


class PredictorEvaluationExecution(Resource['PredictorEvaluationExecution'], Execution):
    """The execution of a PredictorEvaluationWorkflow.

    Possible statuses are INPROGRESS, SUCCEEDED, and FAILED.
    Predictor evaluation executions also have a ``status_description`` field with more information.

    """

    evaluator_names = properties.List(properties.String, "evaluator_names", serializable=False)
    """:List[str]: names of the predictor evaluators that were executed. These are used
    when calling the ``results()`` method."""
    workflow_id = properties.UUID('workflow_id', serializable=False)
    """:UUID: Unique identifier of the workflow that was executed"""
    predictor_id = properties.UUID('predictor_id', serializable=False)
    predictor_version = properties.Integer('predictor_version', serializable=False)

    def _path(self):
        return format_escaped_url(
            '/projects/{project_id}/predictor-evaluation-executions/{execution_id}',
            project_id=self.project_id,
            execution_id=self.uid
        )

    @lru_cache()
    def results(self, evaluator_name: str) -> PredictorEvaluationResult:
        """
        Get a specific evaluation result by the name of the evaluator that produced it.

        Parameters
        ----------
        evaluator_name: str
            Name of the evaluator for which to get the results

        Returns
        -------
        PredictorEvaluationResult
            The evaluation result from the evaluator with the given name

        """
        params = {"evaluator_name": evaluator_name}
        resource = self._session.get_resource(self._path() + "/results", params=params)
        return PredictorEvaluationResult.build(resource)

    def __getitem__(self, item):
        if isinstance(item, str):
            return self.results(item)
        else:
            raise TypeError("Results are accessed by string names")

    def __iter__(self):
        return iter(self.evaluator_names)



"""Top-level class for all attribute template objects and collections thereof."""
from abc import ABC
from typing import TypeVar

from citrine._serialization.properties import Optional as PropertyOptional
from citrine._serialization.properties import String, Object
from gemd.entity.template.attribute_template import AttributeTemplate as GEMDAttributeTemplate
from gemd.entity.bounds.base_bounds import BaseBounds
from citrine.resources.templates import Template, TemplateCollection


class AttributeTemplate(Template, GEMDAttributeTemplate, ABC):
    """
    An abstract attribute template object.

    AttributeTemplate must be extended along with `Resource`
    """

    name = String('name')
    description = PropertyOptional(String(), 'description')
    bounds = Object(BaseBounds, 'bounds', override=True)


AttributeTemplateResourceType = TypeVar("AttributeTemplateResourceType", bound="AttributeTemplate")


class AttributeTemplateCollection(TemplateCollection[AttributeTemplateResourceType]):
    """A collection of one kind of attribute template object."""



"""Tools for working with reports."""
from typing import Type, Dict, TypeVar, Iterable, Any, Set
from abc import abstractmethod
from itertools import groupby
from logging import getLogger

from citrine._serialization import properties
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._serialization.serializable import Serializable
from citrine._rest.asynchronous_object import AsynchronousObject
from citrine.informatics.descriptors import Descriptor
from citrine.informatics.predictor_evaluation_result import ResponseMetrics

SelfType = TypeVar('SelfType', bound='Report')

logger = getLogger(__name__)


class Report(PolymorphicSerializable['Report'], AsynchronousObject):
    """A Citrine Report contains information related to a module.

    Abstract type that returns the proper type given a serialized dict.

    """

    _response_key = None

    @classmethod
    def get_type(cls, data) -> Type[Serializable]:
        """Return the only subtype."""
        return PredictorReport

    @classmethod
    def build(cls, data: dict) -> SelfType:
        """Build the underlying type."""
        subtype = cls.get_type(data)
        report = subtype.build(data)
        report.post_build()
        return report

    @abstractmethod
    def post_build(self):
        """Executes after a .build() is called in [[Report]]."""


class FeatureImportanceReport(Serializable["FeatureImportanceReport"]):
    """Feature importances for a specific model response.

    FeatureImportanceReport objects are constructed from saved models and
    should not be user-instantiated.
    """

    output_key = properties.String('response_key')
    """:str: output descriptor key for which these feature importances are applicable"""
    importances = properties.Mapping(keys_type=properties.String, values_type=properties.Float,
                                     serialization_path='importances')
    """:dict[str, float]: map from feature name to its importance"""

    def __init__(self):
        pass  # pragma: no cover

    def __str__(self):
        return "<FeatureImportanceReport {!r}>".format(self.output_key)  # pragma: no cover


class ModelEvaluationResult(Serializable["ModelEvaluationResult"]):
    """Settings and evaluation metrics for a single algorithm from AutoML model selection.

    ModelEvaluationResult objects are included in a ModelSelectionReport
    and should not be user-instantiated.
    """

    model_settings = properties.Raw('model_settings')
    _response_results = properties.Mapping(
        properties.String,
        properties.Object(ResponseMetrics),
        "response_results"
    )

    def __init__(self):
        pass  # pragma: no cover

    def __str__(self):
        return '<ModelEvaluationResult>'  # pragma: no cover

    def __getitem__(self, item):
        return self._response_results[item]

    def __iter__(self):
        return iter(self.responses)

    @property
    def responses(self) -> Set[str]:
        """Responses the model was evaluated on."""
        return set(self._response_results.keys())


class ModelSelectionReport(Serializable["ModelSelectionReport"]):
    """Summary of selection settings and model performance from AutoML model selection.

    ModelSelectionReport objects are constructed from saved models and
    should not be user-instantiated.
    """

    n_folds = properties.Integer('n_folds')
    evaluation_results = properties.List(
        properties.Object(ModelEvaluationResult),
        "evaluation_results"
    )

    def __init__(self):
        pass  # pragma: no cover

    def __str__(self):
        return '<ModelSelectionReport>'  # pragma: no cover


class ModelSummary(Serializable['ModelSummary']):
    """Summary of information about a single model in a predictor.

    ModelSummary objects are constructed from saved models and should not be user-instantiated.
    """

    name = properties.String('name')
    """:str: the name of the model"""
    type_ = properties.String('type')
    """:str: the type of the model (e.g., "ML Model", "Featurizer", etc.)"""
    inputs = properties.List(
        properties.Union([properties.Object(Descriptor), properties.String()]),
        'inputs'
    )
    """:List[Descriptor]: list of input descriptors"""
    outputs = properties.List(
        properties.Union([properties.Object(Descriptor), properties.String()]),
        'outputs'
    )
    """:List[Descriptor]: list of output descriptors"""
    model_settings = properties.Raw('model_settings')
    """:dict: model settings, as a dictionary (keys depend on the model type)"""
    feature_importances = properties.List(
        properties.Object(FeatureImportanceReport), 'feature_importances')
    """:List[FeatureImportanceReport]: feature importance reports for each output"""
    selection_summary = properties.Optional(
        properties.Object(ModelSelectionReport), "selection_summary"
    )
    """:Optional[ModelSelectionReport]: optional results of AutoML model selection"""
    predictor_name = properties.String('predictor_configuration_name', default='')
    """:str: the name of the predictor that created this model"""
    predictor_uid = properties.Optional(properties.UUID(), 'predictor_configuration_uid')
    """:Optional[UUID]: the unique Citrine id of the predictor that created this model"""
    training_data_count = properties.Optional(properties.Integer, "training_data_count")
    """:int: Number of rows in the training data for the model, if applicable."""

    def __init__(self):
        pass  # pragma: no cover

    def __str__(self):
        return '<ModelSummary {!r}>'.format(self.name)  # pragma: no cover


class PredictorReport(Serializable['PredictorReport'], Report):
    """The performance metrics corresponding to a predictor.

    PredictorReport objects are constructed from saved models and should not be user-instantiated.
    """

    _in_progress_statuses = ["PENDING"]
    _succeeded_statuses = ["OK"]
    _failed_statuses = ["ERROR"]

    uid = properties.Optional(properties.UUID, 'id', serializable=False)
    """:UUID: Unique Citrine id of the predictor report"""
    status = properties.String('status')
    """:str: The status of the report. Possible statuses are PENDING, ERROR, and OK."""
    descriptors = properties.List(properties.Object(Descriptor), 'report.descriptors', default=[])
    """:List[Descriptor]: All descriptors that appear in the predictor"""
    model_summaries = properties.List(properties.Object(ModelSummary), 'report.models', default=[])
    """:List[ModelSummary]: Summaries of all models in the predictor"""

    def __init__(self):
        pass  # pragma: no cover

    def post_build(self):
        """Modify a PredictorReport object in-place after deserialization."""
        self._fill_out_descriptors()
        for _, summary in enumerate(self.model_summaries):
            # Collapse settings on final trained model
            summary.model_settings = self._collapse_model_settings(summary.model_settings)
            if summary.selection_summary is not None:
                # Collapse settings on any child model evaluation results
                for result in summary.selection_summary.evaluation_results:
                    result.model_settings = self._collapse_model_settings(result.model_settings)

    def _fill_out_descriptors(self):
        """Replace descriptor keys in `model_summaries` with full Descriptor objects."""
        descriptor_map = dict()
        for key, group in groupby(sorted(self.descriptors, key=lambda d: d.key), lambda d: d.key):
            descriptor_map[key] = self._get_sole_descriptor(group)
        for i, model in enumerate(self.model_summaries):
            for j, input_key in enumerate(model.inputs):
                try:
                    model.inputs[j] = descriptor_map[input_key]
                except KeyError:
                    raise RuntimeError("Model {} contains input \'{}\', but no descriptor found "
                                       "with that key".format(model.name, input_key))
            for j, output_key in enumerate(model.outputs):
                try:
                    model.outputs[j] = descriptor_map[output_key]
                except KeyError:
                    raise RuntimeError("Model {} contains output \'{}\', but no descriptor found "
                                       "with that key".format(model.name, output_key))

    @staticmethod
    def _get_sole_descriptor(it: Iterable):
        """Get what should be the sole descriptor in an iterable of descriptors.

        This method is called by `_fill_out_descriptors` on an iterable of descriptors
        that have been grouped by their key.

        Parameters
        ----------
        it: Iterable
            iterable of descriptors

        """
        as_list = list(it)
        if len(as_list) > 1:
            serialized_descriptors = [d.dump() for d in as_list]
            logger.warning("Warning: found multiple descriptors with the key \'{}\', arbitrarily "
                           "selecting the first one. The descriptors are: {}"
                           .format(as_list[0].key, serialized_descriptors))
        return as_list[0]

    @staticmethod
    def _collapse_model_settings(raw_settings: Dict[str, Any]):
        """Collapse a model's settings into a flat dictionary.

        Model settings are returned as a dictionary with a "name" field, a "value" field,
        and "children," a list of sub-settings. This method flattens that structure to a
        top-level dictionary with keys given by "name" and values given by "value."

        """
        def _recurse_model_settings(settings: Dict[str, str], list_or_dict):
            """Recursively traverse the model settings, adding name-value pairs to dictionary."""
            if isinstance(list_or_dict, list):
                for setting in list_or_dict:
                    _recurse_model_settings(settings, setting)
            elif isinstance(list_or_dict, dict):
                settings[list_or_dict['name']] = list_or_dict['value']
                _recurse_model_settings(settings, list_or_dict['children'])

        collapsed = dict()
        _recurse_model_settings(collapsed, raw_settings)
        return collapsed



from abc import ABC, abstractmethod
from collections import defaultdict
from typing import List, Iterable

from citrine.resources.data_concepts import DataConcepts

from gemd.util import make_index, writable_sort_order


class Batcher(ABC):
    """Base class for Data Concepts batching routines."""

    @abstractmethod
    def batch(self, objects: Iterable[DataConcepts], batch_size) -> List[List[DataConcepts]]:
        """Collect a list of DataConcepts into batches according to some batching algorithm."""

    @staticmethod
    def by_type():
        """Return a BatchByType batcher."""
        return BatchByType()

    @staticmethod
    def by_dependency():
        """Return a BatchByDependency batcher."""
        return BatchByDependency()


class BatchByType(Batcher):
    """Batching by object type."""

    def batch(self, objects: Iterable[DataConcepts], batch_size) -> List[List[DataConcepts]]:
        """Collect object batches by type, following an order that will satisfy prereqs."""
        batches = list()
        by_type = defaultdict(list)
        seen = {}
        for obj in objects:
            if obj.to_link() in seen:  # Repeat in the iterable; don't add it to the batch
                if seen[obj.to_link()] != obj:  # verify that it's a replicate
                    raise ValueError(f"Colliding objects for {obj.to_link()}")
            else:
                by_type[obj.typ].append(obj)
                for scope in obj.uids:
                    seen[obj.to_link(scope)] = obj
        typ_groups = sorted(list(by_type.values()), key=lambda x: writable_sort_order(x[0]))
        for typ_group in typ_groups:
            num_batches = len(typ_group) // batch_size
            for batch_num in range(num_batches + 1):
                batch = typ_group[batch_num * batch_size: (batch_num + 1) * batch_size]
                batches.append(batch)
        for i in reversed(range(len(batches) - 1)):
            if len(batches[i]) + len(batches[i + 1]) <= batch_size:
                batches[i].extend(batches[i + 1])
                del batches[i + 1]

        return batches


class BatchByDependency(Batcher):
    """Batching by clusters where nothing references anything outside the cluster."""

    def batch(self, objects: Iterable[DataConcepts], batch_size) -> List[List[DataConcepts]]:
        """Collect object batches that are internally consistent for dry_run object tests."""
        # Collect shallow dependences, UID references, and type-based clusters
        depends = dict()
        obj_set = set(objects)
        index = make_index(objects)
        by_type = defaultdict(list)
        for obj in objects:  # Don't worry about replicates since we'd have them anyway
            depends[obj] = obj.all_dependencies()
            by_type[obj.typ].append(obj)

        # Deep dependencies w/ objects only, build inverse index
        # This takes a second loop because we need to build up all derefs first
        supported_by = defaultdict(list)
        type_groups = sorted(list(by_type.values()), key=lambda x: writable_sort_order(x[0]))
        for type_group in type_groups:
            for obj in type_group:
                # Collect objects of interest that we are supposed to load
                # Note depends contains both objects & links; obj_set is everything in the call
                local_set = {index.get(x, x) for x in depends[obj] if index.get(x, x) in obj_set}
                full_set = set(local_set)
                if len(full_set) > batch_size:
                    raise ValueError(f"Object {obj.name} has more than {batch_size} dependencies.")

                for subobj in local_set:
                    full_set.update(depends[subobj])
                depends[obj] = sorted(list(full_set),
                                      key=lambda x: writable_sort_order(x))
                for dependant in reversed(depends[obj]):
                    supported_by[dependant].append(obj)

        # Build self-consistent clusters
        queued = set()
        clusters = list()
        for type_group in reversed(type_groups):  # Later objects have more dependencies
            for obj in type_group:
                if obj in queued:
                    continue  # It's already in a cluster
                to_be_checked = depends[obj]
                cluster = {obj} | set(depends[obj])
                while to_be_checked:
                    parent = to_be_checked.pop()
                    for candidate in reversed(supported_by[parent]):
                        if candidate in queued or candidate in cluster:
                            continue  # It's already in a cluster
                        candidate_set = {candidate} | set(depends[candidate])
                        if len(cluster | candidate_set) > batch_size:
                            continue  # It wouldn't fit

                        to_be_checked.extend(candidate_set - cluster)
                        cluster |= candidate_set

                clusters.append(list(cluster))
                queued.update(cluster)

        return clusters



from functools import partial
from typing import Iterable

from citrine._rest.paginator import Paginator
from citrine._rest.resource import Resource
from citrine._serialization import properties
from citrine._utils.functions import format_escaped_url
from citrine.informatics.descriptors import Descriptor
from citrine.informatics.design_candidate import DesignCandidate
from citrine.informatics.predict_request import PredictRequest
from citrine.informatics.scores import Score
from citrine.informatics.executions.execution import Execution


class DesignExecution(Resource["DesignExecution"], Execution):
    """The execution of a DesignWorkflow.

    Possible statuses are INPROGRESS, SUCCEEDED, and FAILED.
    Design executions also have a ``status_description`` field with more information.

    """

    _paginator: Paginator = Paginator()
    _collection_key = 'response'
    workflow_id = properties.UUID('workflow_id', serializable=False)
    """:UUID: Unique identifier of the workflow that was executed"""
    version_number = properties.Integer("version_number", serializable=False)
    """:int: Integer identifier that increases each time the workflow is executed. The first
    execution has version_number = 1."""

    score = properties.Object(Score, 'score')
    """:Score: score by which this execution was evaluated"""
    descriptors = properties.List(properties.Object(Descriptor), 'descriptors')
    """:List[Descriptor]: all of the descriptors in the candidates generated by this execution"""

    def _path(self):
        return format_escaped_url(
            '/projects/{project_id}/design-workflows/{workflow_id}/executions/{execution_id}',
            project_id=self.project_id,
            workflow_id=self.workflow_id,
            execution_id=self.uid
        )

    @classmethod
    def _build_candidates(cls, subset_collection: Iterable[dict]) -> Iterable[DesignCandidate]:
        for candidate in subset_collection:
            yield DesignCandidate.build(candidate)

    def candidates(self, *, per_page: int = 100) -> Iterable[DesignCandidate]:
        """Fetch the Design Candidates for the particular execution, paginated."""
        path = self._path() + '/candidates'

        fetcher = partial(self._fetch_page, path=path, fetch_func=self._session.get_resource)

        return self._paginator.paginate(page_fetcher=fetcher,
                                        collection_builder=self._build_candidates,
                                        per_page=per_page)

    def predict(self,
                predict_request: PredictRequest) -> DesignCandidate:
        """Invoke a prediction on a design candidate."""
        path = self._path() + '/predict'

        res = self._session.post_resource(path, predict_request.dump(), version=self._api_version)
        return DesignCandidate.build(res)



from citrine._serialization import properties
from citrine._serialization.serializable import Serializable


class ValidationError(Serializable["ValidationError"]):
    """A user-facing error message describing why their request was invalid."""

    failure_message = properties.Optional(properties.String(), "failure_message")
    property = properties.Optional(properties.String(), "property")
    input = properties.Optional(properties.String(), "input")
    failure_id = properties.Optional(properties.String(), "failure_id")


class ApiError(Serializable["ApiError"]):
    """The engineering API root level error model."""

    code = properties.Optional(properties.Integer(), "code")
    message = properties.Optional(properties.String(), "message")
    validation_errors = properties.List(properties.Object(ValidationError), "validation_errors")

    def has_failure(self, failure_id: str) -> bool:
        """Checks if this error contains a ValidationError with specified failure ID."""
        if not failure_id:
            # Discourage both anonymous errors ('') and confusing has_failure(None)
            # syntax which may imply there is no failure at all.
            raise ValueError(f"failure_id cannot be empty: '{failure_id}'")
        return any(v.failure_id == failure_id for v in self.validation_errors)



"""Tools for working with module resources."""
from typing import Type, Optional
from uuid import UUID

from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._session import Session
from citrine._rest.asynchronous_object import AsynchronousObject

__all__ = ['Module']


class Module(PolymorphicSerializable['Module'], AsynchronousObject):
    """A Citrine Module is a reusable computational tool used to construct a workflow.

    Abstract type that returns the proper type given a serialized dict.

    All modules must inherit AIResourceMetadata, and hence have a ``status`` field.
    Possible statuses are CREATED, VALIDATING, INVALID, ERROR, and READY.

    """

    _response_key = None
    _project_id: Optional[UUID] = None
    _session: Optional[Session] = None
    _in_progress_statuses = ["VALIDATING", "CREATED"]
    _succeeded_statuses = ["READY"]
    _failed_statuses = ["INVALID", "ERROR"]

    @classmethod
    def get_type(cls, data) -> Type['Module']:
        """Return the subtype."""
        from citrine.informatics.design_spaces import DesignSpace
        return DesignSpace



from typing import Optional, Union
from uuid import UUID

from citrine._rest.resource import Resource
from citrine._serialization import properties
from citrine.informatics.workflows.workflow import Workflow
from citrine.resources.design_execution import DesignExecutionCollection
from citrine._rest.ai_resource_metadata import AIResourceMetadata

__all__ = ['DesignWorkflow']


class DesignWorkflow(Resource['DesignWorkflow'], Workflow, AIResourceMetadata):
    """Object that generates scored materials that may approach higher values of the score.

    Parameters
    ----------
    name: str
        the name of the workflow
    design_space_id: Optional[UUID]
        the UUID corresponding to the design space to use
    predictor_id: Optional[UUID]
        the UUID corresponding to the predictor to use
    predictor_version: Optional[Union[int, str]]
        the version of the predictor to use
    description: Optional[str]
        a description of the workflow

    """

    design_space_id = properties.Optional(properties.UUID, 'design_space_id')
    predictor_id = properties.Optional(properties.UUID, 'predictor_id')
    predictor_version = properties.Optional(
        properties.Union([properties.Integer(), properties.String()]), 'predictor_version')
    branch_id: Optional[UUID] = properties.Optional(properties.UUID, 'branch_id')
    """:Optional[UUID]: Unique ID of the branch that contains this workflow."""

    status_description = properties.String('status_description', serializable=False)
    """:str: more detailed description of the workflow's status"""
    typ = properties.String('type', default='DesignWorkflow', deserializable=False)

    def __init__(self,
                 name: str,
                 *,
                 design_space_id: Optional[UUID] = None,
                 predictor_id: Optional[UUID] = None,
                 predictor_version: Optional[Union[int, str]] = None,
                 description: Optional[str] = None):
        self.name = name
        self.design_space_id = design_space_id
        self.predictor_id = predictor_id
        self.predictor_version = predictor_version
        self.description = description

    def __str__(self):
        return '<DesignWorkflow {!r}>'.format(self.name)

    @property
    def design_executions(self) -> DesignExecutionCollection:
        """Return a resource representing all visible executions of this workflow."""
        if getattr(self, 'project_id', None) is None:
            raise AttributeError('Cannot initialize execution without project reference!')
        return DesignExecutionCollection(
            project_id=self.project_id, session=self._session, workflow_id=self.uid)



"""Resources that represent process spec objects."""
from typing import Optional, Dict, List, Type, Union, Iterator
from uuid import UUID

from citrine._rest.resource import GEMDResource
from citrine._serialization.properties import List as PropertyList
from citrine._serialization.properties import Optional as PropertyOptional
from citrine._serialization.properties import String, Object, LinkOrElse
from citrine.resources.object_specs import ObjectSpec, ObjectSpecCollection
from gemd.entity.attribute.condition import Condition
from gemd.entity.attribute.parameter import Parameter
from gemd.entity.file_link import FileLink
from gemd.entity.link_by_uid import LinkByUID
from gemd.entity.object.process_spec import ProcessSpec as GEMDProcessSpec
from gemd.entity.template.process_template import ProcessTemplate as GEMDProcessTemplate


class ProcessSpec(
    GEMDResource['ProcessSpec'],
    ObjectSpec,
    GEMDProcessSpec,
    typ=GEMDProcessSpec.typ
):
    """
    A process specification.

    Processes transform zero or more input materials into exactly one output material.

    Parameters
    ----------
    name: str
        Name of the process spec.
    uids: Map[str, str], optional
        A collection of
        `unique IDs <https://citrineinformatics.github.io/gemd-docs/
        specification/unique-identifiers/>`_.
    tags: List[str], optional
        `Tags <https://citrineinformatics.github.io/gemd-docs/specification/tags/>`_
        are hierarchical strings that store information about an entity. They can be used
        for filtering and discoverability.
    notes: str, optional
        Long-form notes about the process spec.
    conditions: List[Condition], optional
        Conditions under which this process spec occurs.
    parameters: List[Parameter], optional
        Parameters of this process spec.
    template: ProcessTemplate, optional
        A template bounding the valid values for this process's parameters and conditions.
    file_links: List[FileLink], optional
        Links to associated files, with resource paths into the files API.

    Attributes
    ----------
    output_material: MaterialSpec
        The material spec that this process spec produces. The link is established by creating
        the material spec and settings its `process` field to this process spec.

    ingredients: List[IngredientSpec], optional
        Ingredient specs that act as inputs to this process spec. The link is established by
        creating each ingredient spec and setting its `process` field to this process spec.

    """

    _response_key = GEMDProcessSpec.typ  # 'process_spec'

    name = String('name', override=True, use_init=True)
    conditions = PropertyOptional(PropertyList(Object(Condition)), 'conditions', override=True)
    parameters = PropertyOptional(PropertyList(Object(Parameter)), 'parameters', override=True)
    template = PropertyOptional(LinkOrElse(GEMDProcessTemplate),
                                'template', override=True,
                                use_init=True,
                                )

    def __init__(self,
                 name: str,
                 *,
                 uids: Optional[Dict[str, str]] = None,
                 tags: Optional[List[str]] = None,
                 notes: Optional[str] = None,
                 conditions: Optional[List[Condition]] = None,
                 parameters: Optional[List[Parameter]] = None,
                 template: Optional[GEMDProcessTemplate] = None,
                 file_links: Optional[List[FileLink]] = None
                 ):
        if uids is None:
            uids = dict()
        super(ObjectSpec, self).__init__()
        GEMDProcessSpec.__init__(self, name=name, uids=uids,
                                 tags=tags, conditions=conditions, parameters=parameters,
                                 template=template, file_links=file_links, notes=notes)

    def __str__(self):
        return '<Process spec {!r}>'.format(self.name)


class ProcessSpecCollection(ObjectSpecCollection[ProcessSpec]):
    """Represents the collection of all process specs associated with a dataset."""

    _path_template = 'projects/{project_id}/datasets/{dataset_id}/process-specs'
    _dataset_agnostic_path_template = 'projects/{project_id}/process-specs'
    _individual_key = 'process_spec'
    _collection_key = 'process_specs'
    _resource = ProcessSpec

    @classmethod
    def get_type(cls) -> Type[ProcessSpec]:
        """Return the resource type in the collection."""
        return ProcessSpec

    def list_by_template(self,
                         uid: Union[UUID, str, LinkByUID, GEMDProcessTemplate]
                         ) -> Iterator[ProcessSpec]:
        """
        Get the process specs using the specified process template.

        Parameters
        ----------
        uid: Union[UUID, str, LinkByUID, GEMDProcessTemplate]
            A representation of the process template whose process spec usages are to be located.

        Returns
        -------
        Iterator[ProcessSpec]
            The process specs using the specified process template

        """
        return self._get_relation('process-templates', uid=uid)



__all__ = ['Predictor']


class Predictor:
    """Parent type of the individual predictors on the Citrine Platform.

    The Predictor class unifies the composite GraphPredictor and individual PredictorNode types
    but does not provide any functionality on its own.

    """



import warnings
from datetime import datetime
from typing import Type, Optional, List
from uuid import UUID

from deprecation import deprecated

from citrine._serialization import properties
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine.informatics.data_sources import DataSource
from citrine.informatics.predictors import Predictor
from citrine.resources.status_detail import StatusDetail


def _check_deprecated_training_data(training_data: Optional[List[DataSource]]) -> None:
    if training_data is not None:
        warnings.warn(
            f"The field `training_data` on single predictor nodes is deprecated "
            "and will be removed in version 3.0.0. Include training data for all "
            "sub-predictors on the parent GraphPredictor. Existing training data "
            "on this predictor will be moved to the parent GraphPredictor upon registration.",
            DeprecationWarning
        )


class PredictorNode(PolymorphicSerializable["PredictorNode"], Predictor):
    """An individual compute node within a Predictor.

    A PredictorNode cannot be registered to the Citrine Platform by itself
    and must be included as a component within a GraphPredictor to be used.

    """

    name = properties.String("name")
    description = properties.Optional(properties.String(), "description")

    @classmethod
    def get_type(cls, data) -> Type['PredictorNode']:
        """Return the subtype."""
        from .expression_predictor import ExpressionPredictor
        from .molecular_structure_featurizer import MolecularStructureFeaturizer
        from .ingredients_to_formulation_predictor import IngredientsToFormulationPredictor
        from .label_fractions_predictor import LabelFractionsPredictor
        from .simple_mixture_predictor import SimpleMixturePredictor
        from .ingredient_fractions_predictor import IngredientFractionsPredictor
        from .auto_ml_predictor import AutoMLPredictor
        from .mean_property_predictor import MeanPropertyPredictor
        from .chemical_formula_featurizer import ChemicalFormulaFeaturizer
        type_dict = {
            "AnalyticExpression": ExpressionPredictor,
            "MoleculeFeaturizer": MolecularStructureFeaturizer,
            "IngredientsToSimpleMixture": IngredientsToFormulationPredictor,
            "MeanProperty": MeanPropertyPredictor,
            "LabelFractions": LabelFractionsPredictor,
            "SimpleMixture": SimpleMixturePredictor,
            "IngredientFractions": IngredientFractionsPredictor,
            "ChemicalFormulaFeaturizer": ChemicalFormulaFeaturizer,
            "AutoML": AutoMLPredictor,
        }
        typ = type_dict.get(data['type'])
        if typ is not None:
            return typ
        else:
            raise ValueError(
                '{} is not a valid predictor node type. '
                'Must be in {}.'.format(data['type'], type_dict.keys())
            )

    @property
    @deprecated(
        deprecated_in="2.13.0",
        removed_in="3.0.0",
        details="Use `uid` on parent GraphPredictor."
    )
    def uid(self) -> Optional[UUID]:
        """Optional Citrine Platform unique identifier."""
        return None

    @property
    @deprecated(
        deprecated_in="2.13.0",
        removed_in="3.0.0",
        details="Use `version` on parent GraphPredictor."
    )
    def version(self) -> Optional[int]:
        """The version number of the resource."""
        return None

    @property
    @deprecated(
        deprecated_in="2.13.0",
        removed_in="3.0.0",
        details="Use `draft` on parent GraphPredictor."
    )
    def draft(self) -> Optional[bool]:
        """The draft status of the resource."""
        return None

    @property
    @deprecated(
        deprecated_in="2.13.0",
        removed_in="3.0.0",
        details="Use `created_by` on parent GraphPredictor."
    )
    def created_by(self) -> Optional[UUID]:
        """The id of the user who created the resource."""
        return None

    @property
    @deprecated(
        deprecated_in="2.13.0",
        removed_in="3.0.0",
        details="Use `updated_by` on parent GraphPredictor."
    )
    def updated_by(self) -> Optional[UUID]:
        """The id of the user who most recently updated the resource."""
        return None

    @property
    @deprecated(
        deprecated_in="2.13.0",
        removed_in="3.0.0",
        details="Use `archived_by` on parent GraphPredictor."
    )
    def archived_by(self) -> Optional[UUID]:
        """The id of the user who most recently archived the resource."""
        return None

    @property
    @deprecated(
        deprecated_in="2.13.0",
        removed_in="3.0.0",
        details="Use `create_time` on parent GraphPredictor."
    )
    def create_time(self) -> Optional[datetime]:
        """The date and time at which the resource was created."""
        return None

    @property
    @deprecated(
        deprecated_in="2.13.0",
        removed_in="3.0.0",
        details="Use `update_time` on parent GraphPredictor."
    )
    def update_time(self) -> Optional[datetime]:
        """The date and time at which the resource was most recently updated."""
        return None

    @property
    @deprecated(
        deprecated_in="2.13.0",
        removed_in="3.0.0",
        details="Use `archive_time` on parent GraphPredictor."
    )
    def archive_time(self) -> Optional[datetime]:
        """The date and time at which the resource was archived."""
        return None

    @property
    @deprecated(
        deprecated_in="2.13.0",
        removed_in="3.0.0",
        details="Check `status` on parent GraphPredictor."
    )
    def status(self) -> Optional[str]:
        """Short description of the resource's status."""
        return None

    @property
    @deprecated(
        deprecated_in="2.13.0",
        removed_in="3.0.0",
        details="Check `status_detail` on parent GraphPredictor."
    )
    def status_detail(self) -> List[StatusDetail]:
        """A list of structured status info, containing the message and level."""
        return []

    @deprecated(
        deprecated_in="2.13.0",
        removed_in="3.0.0",
        details="Check `in_progress` on parent GraphPredictor."
    )
    def in_progress(self) -> bool:
        """Whether the backend process is in progress."""
        return False

    @deprecated(
        deprecated_in="2.13.0",
        removed_in="3.0.0",
        details="Check `succeeded` on parent GraphPredictor."
    )
    def succeeded(self) -> bool:
        """Whether the backend process has completed successfully."""
        return False

    @deprecated(
        deprecated_in="2.13.0",
        removed_in="3.0.0",
        details="Check `failed` on parent GraphPredictor."
    )
    def failed(self) -> bool:
        """Whether the backend process has completed unsuccessfully."""
        return False



"""Tools for working with design spaces."""
from typing import Optional, Type
from uuid import UUID

from citrine._rest.asynchronous_object import AsynchronousObject
from citrine._serialization import properties
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._serialization.serializable import Serializable
from citrine._session import Session
from citrine.resources.sample_design_space_execution import \
    SampleDesignSpaceExecutionCollection

from deprecation import deprecated


__all__ = ['DesignSpace']


class DesignSpace(PolymorphicSerializable['DesignSpace'], AsynchronousObject):
    """A Citrine Design Space describes the set of materials that can be made.

    Abstract type that returns the proper type given a serialized dict.

    """

    @property
    @deprecated(deprecated_in="2.26.0", removed_in="3.0.0",
                details="Please use `isinstance` or `issubclass` instead.")
    def module_type(self):
        """The type of module."""
        return "DESIGN_SPACE"

    @module_type.setter
    @deprecated(deprecated_in="2.26.0", removed_in="3.0.0")
    def module_type(self, value):
        pass

    uid = properties.Optional(properties.UUID, 'id', serializable=False)
    """:Optional[UUID]: Citrine Platform unique identifier"""
    name = properties.String('data.name')
    description = properties.Optional(properties.String(), 'data.description')

    @staticmethod
    def wrap_instance(subspace_data: dict) -> dict:
        """Insert a serialized embedded design space into an entity envelope.

        This facilitates deserialization.
        """
        return {
            "data": {
                "name": subspace_data.get("name", ""),
                "description": subspace_data.get("description", ""),
                "instance": subspace_data
            }
        }

    _response_key = None
    _project_id: Optional[UUID] = None
    _session: Optional[Session] = None
    _in_progress_statuses = ["VALIDATING", "CREATED"]
    _succeeded_statuses = ["READY"]
    _failed_statuses = ["INVALID", "ERROR"]

    @classmethod
    def get_type(cls, data) -> Type[Serializable]:
        """Return the subtype."""
        from .data_source_design_space import DataSourceDesignSpace
        from .enumerated_design_space import EnumeratedDesignSpace
        from .formulation_design_space import FormulationDesignSpace
        from .product_design_space import ProductDesignSpace
        from .hierarchical_design_space import HierarchicalDesignSpace

        return {
            'Univariate': ProductDesignSpace,
            'ProductDesignSpace': ProductDesignSpace,
            'EnumeratedDesignSpace': EnumeratedDesignSpace,
            'FormulationDesignSpace': FormulationDesignSpace,
            'DataSourceDesignSpace': DataSourceDesignSpace,
            'HierarchicalDesignSpace': HierarchicalDesignSpace
        }[data['data']['instance']['type']]

    @property
    def sample_design_space_executions(self):
        """Start a Sample Design Space Execution using the current Design Space."""
        return SampleDesignSpaceExecutionCollection(
            project_id=self._project_id, design_space_id=self.uid, session=self._session
        )



from typing import TypeVar, Union

from citrine._serialization.serializable import Serializable
from citrine._serialization import properties

from gemd.enumeration.base_enumeration import BaseEnumeration


StatusDetailType = TypeVar('StatusDetailType', bound='StatusDetail')


class StatusLevelEnum(BaseEnumeration):
    """The level of the status message."""

    INFO = "Info"
    WARNING = "Warning"
    ERROR = "Error"
    UNKNOWN = "Unknown"


class StatusDetail(Serializable[StatusDetailType]):
    """A status message and its level."""

    msg = properties.String("msg")
    level = properties.String("level")

    def __init__(self, *, msg: str, level: Union[str, StatusLevelEnum]):
        self.msg = msg
        self.level = StatusLevelEnum.from_str(level, exception=True)

    def __str__(self):
        return f"[{self.level.upper()}] {self.msg}"

    def __repr__(self):
        return f"<{self}>"



from abc import abstractmethod
from logging import getLogger
from typing import Optional, Union, Generic, TypeVar, Iterable, Iterator, Sequence, Dict
from urllib.parse import urlparse, urlencode
from uuid import UUID

from citrine._rest.pageable import Pageable
from citrine._rest.paginator import Paginator
from citrine._rest.resource import ResourceRef
from citrine._utils.functions import format_escaped_url
from citrine.exceptions import ModuleRegistrationFailedException, NonRetryableException
from citrine.resources.response import Response

logger = getLogger(__name__)

ResourceType = TypeVar('ResourceType', bound='Resource')

# Python does not support a TypeVar being used as a bound for another TypeVar.
# Thus, this will never be particularly type safe on its own. The solution is to
# have subclasses override the create method.
CreationType = TypeVar('CreationType', bound='Resource')


class Collection(Generic[ResourceType], Pageable):
    """Abstract class for representing collections of REST resources."""

    _path_template: str = NotImplemented
    _dataset_agnostic_path_template: str = NotImplemented
    _individual_key: str = NotImplemented
    _resource: ResourceType = NotImplemented
    _collection_key: str = 'entries'
    _paginator: Paginator = Paginator()
    _api_version: str = "v1"

    def _put_resource_ref(self, subpath: str, uid: Union[UUID, str]):
        url = self._get_path(subpath)
        ref = ResourceRef(uid)
        return self.session.put_resource(url, ref.dump(), version=self._api_version)

    def _get_path(self,
                  uid: Optional[Union[UUID, str]] = None,
                  *,
                  ignore_dataset: bool = False,
                  action: Union[str, Sequence[str]] = [],
                  query_terms: Dict[str, str] = {},
                  ) -> str:
        """Construct a url from __base_path__ and, optionally, id and/or action."""
        if ignore_dataset:
            base = urlparse(self._dataset_agnostic_path_template)
        else:
            base = urlparse(self._path_template)
        path = base.path.split('/')

        if uid is not None:
            path.append("{uid}")

        if isinstance(action, str):
            action = [action]
        else:
            action = list(action)
        path.extend(["{}"] * len(action))

        query = urlencode(query_terms)
        new_url = base._replace(path='/'.join(path), query=query).geturl()

        return format_escaped_url(new_url, *action, **self.__dict__, uid=uid)

    @abstractmethod
    def build(self, data: dict):
        """Build an individual element of the collection."""

    def get(self, uid: Union[UUID, str]) -> ResourceType:
        """Get a particular element of the collection."""
        if uid is None:
            raise ValueError("Cannot get when uid=None.  Are you using a registered resource?")
        path = self._get_path(uid)
        data = self.session.get_resource(path, version=self._api_version)
        data = data[self._individual_key] if self._individual_key else data
        return self.build(data)

    def register(self, model: CreationType) -> CreationType:
        """Create a new element of the collection by registering an existing resource."""
        path = self._get_path()
        try:
            data = self.session.post_resource(path, model.dump(), version=self._api_version)
            data = data[self._individual_key] if self._individual_key else data
            return self.build(data)
        except NonRetryableException as e:
            raise ModuleRegistrationFailedException(model.__class__.__name__, e)

    def list(self, *, per_page: int = 100) -> Iterator[ResourceType]:
        """
        Paginate over the elements of the collection.

        Leaving page and per_page as default values will yield all elements in the
        collection, paginating over all available pages.

        Parameters
        ---------
        per_page: int, optional
            Max number of results to return per page. Default is 100.  This parameter
            is used when making requests to the backend service.  If the page parameter
            is specified it limits the maximum number of elements in the response.

        Returns
        -------
        Iterator[ResourceType]
            An iterator that can be used to loop over all the resources in this collection.
            Use list() to force evaluation of all results into an in-memory list.

        """
        return self._paginator.paginate(page_fetcher=self._fetch_page,
                                        collection_builder=self._build_collection_elements,
                                        per_page=per_page)

    def update(self, model: CreationType) -> CreationType:
        """Update a particular element of the collection."""
        url = self._get_path(model.uid)
        updated = self.session.put_resource(url, model.dump(), version=self._api_version)
        data = updated[self._individual_key] if self._individual_key else updated
        return self.build(data)

    def delete(self, uid: Union[UUID, str]) -> Response:
        """Delete a particular element of the collection."""
        url = self._get_path(uid)
        data = self.session.delete_resource(url, version=self._api_version)
        return Response(body=data)

    def _build_collection_elements(self,
                                   collection: Iterable[dict]) -> Iterator[ResourceType]:
        """
        For each element in the collection, build the appropriate resource type.

        Parameters
        ---------
        collection: Iterable[dict]
            collection containing the elements to be built

        Returns
        -------
        Iterator[ResourceType]
            Resources in this collection.

        """
        for element in collection:
            yield self.build(element)



import warnings
from typing import List, Optional, Union
from uuid import UUID

from citrine._rest.asynchronous_object import AsynchronousObject
from citrine._rest.engine_resource import VersionedEngineResource
from citrine._serialization import properties as properties
from citrine._session import Session
from citrine._utils.functions import format_escaped_url
from citrine.informatics.data_sources import DataSource
from citrine.informatics.predictors.single_predict_request import SinglePredictRequest
from citrine.informatics.predictors.single_prediction import SinglePrediction
from citrine.informatics.predictors import PredictorNode, Predictor
from citrine.resources.report import ReportResource

__all__ = ['GraphPredictor']


class GraphPredictor(VersionedEngineResource['GraphPredictor'], AsynchronousObject, Predictor):
    """A predictor interface that stitches individual predictor nodes together.

    The GraphPredictor is the only predictor that can be registered on the Citrine Platform
    and carries along its meta-data regarding versioning, platform identifiers, and updates.

    Parameters
    ----------
    name: str
        name of the configuration
    description: str
        the description of the predictor
    predictors: List[Union[UUID, PredictorNode]],
        the list of individual predictors to use in the graph
    training_data: Optional[List[DataSource]]
        Optional sources of training data shared by all predictors in the graph.
        Training data provided by this graph predictor does not need to be specified as part of the
        configuration of sub-predictors. Shared training data and any training data specified
        by a sub-predictor will be combined into a flattened list and de-duplicated
        by uid and identifiers. De-duplication is performed if a uid or identifier is shared
        between two or more rows. The content of a de-duplicated row will contain the union of
        data across all rows that share the same uid or at least 1 identifier.

    """

    uid = properties.Optional(properties.UUID, 'id', serializable=False)
    """:Optional[UUID]: Citrine Platform unique identifier"""

    name = properties.String('data.name')
    description = properties.Optional(properties.String(), 'data.description')
    predictors = properties.List(properties.Object(PredictorNode), 'data.instance.predictors')

    # the default seems to be defined in instances, not the class itself
    # this is tested in test_graph_default_training_data
    training_data = properties.List(
        properties.Object(DataSource), 'data.instance.training_data', default=[]
    )

    version = properties.Optional(
        properties.Union([properties.Integer(), properties.String()]),
        'metadata.version',
        serializable=False
    )

    _api_version = "v3"
    _response_key = None
    _project_id: Optional[UUID] = None
    _session: Optional[Session] = None
    _in_progress_statuses = ["VALIDATING", "CREATED"]
    _succeeded_statuses = ["READY"]
    _failed_statuses = ["INVALID", "ERROR"]

    def __init__(self,
                 name: str,
                 *,
                 description: str,
                 predictors: List[Union[UUID, PredictorNode]],
                 training_data: Optional[List[DataSource]] = None):
        self.name: str = name
        self.description: str = description
        self.training_data: List[DataSource] = training_data or []

        uid_predictors = [x for x in predictors if isinstance(x, UUID)]
        if len(uid_predictors) > 0:
            warnings.warn(
                "Referencing predictors by a UUID inside a GraphPredictor is no longer supported "
                "on the Citrine Platform. Please remove all references to predictors "
                "and add only PredictorNode objects to the `predictors` field.",
                DeprecationWarning
            )

        self.predictors: List[PredictorNode] = [
            x for x in predictors if isinstance(x, PredictorNode)
        ]

    def __str__(self):
        return '<GraphPredictor {!r}>'.format(self.name)

    def _path(self):
        return format_escaped_url(
            '/projects/{project_id}/predictors/{predictor_id}/versions/{version}',
            project_id=self._project_id,
            predictor_id=str(self.uid),
            version=self.version
        )

    @staticmethod
    def wrap_instance(predictor_data: dict) -> dict:
        """Insert a serialized embedded predictor into a module envelope.

        This facilitates deserialization.
        """
        return {
            "data": {
                "name": predictor_data.get("name", ""),
                "description": predictor_data.get("description", ""),
                "instance": predictor_data
            }
        }

    @property
    def report(self):
        """Fetch the predictor report."""
        if self.uid is None or self._session is None or self._project_id is None \
                or getattr(self, "version", None) is None:
            msg = "Cannot get the report for a predictor that wasn't read from the platform"
            raise ValueError(msg)
        report_resource = ReportResource(self._project_id, self._session)
        return report_resource.get(predictor_id=self.uid, predictor_version=self.version)

    def predict(self, predict_request: SinglePredictRequest) -> SinglePrediction:
        """Make a one-off prediction with this predictor."""
        path = self._path() + '/predict'
        res = self._session.post_resource(path, predict_request.dump(), version=self._api_version)
        return SinglePrediction.build(res)



"""A collection of FileLink objects."""
from deprecation import deprecated
import mimetypes
import os
from pathlib import Path
from logging import getLogger
from tempfile import TemporaryDirectory
from typing import Optional, Tuple, Union, List, Dict, Iterable, Sequence
from urllib.parse import urlparse, unquote_plus
from uuid import UUID

import requests
from boto3 import client as boto3_client
from boto3.session import Config
from botocore.exceptions import ClientError
from citrine._rest.collection import Collection
from citrine._rest.resource import GEMDResource
from citrine._serialization import properties
from citrine._serialization.serializable import Serializable
from citrine._session import Session
from citrine._utils.functions import rewrite_s3_links_locally
from citrine._utils.functions import write_file_locally

from citrine.jobs.job import JobSubmissionResponse, _poll_for_job_completion
from citrine.resources.response import Response
from gemd.entity.dict_serializable import DictSerializableMeta
from gemd.entity.bounds.base_bounds import BaseBounds
from gemd.entity.file_link import FileLink as GEMDFileLink
from gemd.enumeration.base_enumeration import BaseEnumeration

logger = getLogger(__name__)


class SearchFileFilterTypeEnum(BaseEnumeration):
    """
    The type of the filter used to search for files.

    * SEARCH_BY_NAME:
        Search a file by name in a specific dataset,
        returns by default the last version or a specific one
    * SEARCH_BY_VERSION_ID:
        Search by a specific file version id
    * SEARCH_BY_DATASET_FILE_ID:
        Search either the last version or a specific version number for a specific dataset file id

    """

    NAME_SEARCH = "search_by_name"
    VERSION_ID_SEARCH = "search_by_version_id"
    DATASET_FILE_ID_SEARCH = "search_by_dataset_file_id"


class _Uploader:
    """Holds the many parameters that are generated and used during file upload."""

    def __init__(self):
        self.bucket = ''
        self.object_key = ''
        self.upload_id = ''
        self.region_name = ''
        self.aws_access_key_id = ''
        self.aws_secret_access_key = ''
        self.aws_session_token = ''
        self.s3_version = ''
        self.s3_endpoint_url = None
        self.s3_use_ssl = True
        self.s3_addressing_style = 'auto'


class FileProcessingType(BaseEnumeration):
    """The supported File Processing Types."""

    VALIDATE_CSV = "VALIDATE_CSV"


class FileProcessingData:
    """The base class of all File Processing related data implementations."""

    pass


class CsvColumnInfo(Serializable):
    """The info for a CSV Column, contains the name, recommended and exact bounds."""

    name = properties.String('name')
    """:str: name of the column"""
    bounds = properties.Object(BaseBounds, 'bounds')
    """:BaseBounds: recommended bounds of the column (might include some padding)"""
    exact_range_bounds = properties.Object(BaseBounds, 'exact_range_bounds')
    """:BaseBounds: exact bounds of the column"""

    def __init__(self, name: str, bounds: BaseBounds,
                 exact_range_bounds: BaseBounds):  # pragma: no cover
        self.name = name
        self.bounds = bounds
        self.exact_range_bounds = exact_range_bounds


class CsvValidationData(FileProcessingData, Serializable):
    """The resulting data from the processed CSV file."""

    columns = properties.Optional(properties.List(properties.Object(CsvColumnInfo)),
                                  'columns')
    """:Optional[List[CsvColumnInfo]]: all of the columns in the CSV"""
    record_count = properties.Integer('record_count')
    """:int: the number of rows in the CSV"""

    def __init__(self, columns: List[CsvColumnInfo],
                 record_count: int):  # pragma: no cover
        self.columns = columns
        self.record_count = record_count


class FileProcessingResult:
    """
    The results of a successful file processing operation.

    The type of the actual data depends on the specific processing type.
    """

    def __init__(self,
                 processing_type: FileProcessingType,
                 data: Union[Dict, FileProcessingData]):
        self.processing_type = processing_type
        self.data = data


class FileLinkMeta(DictSerializableMeta):
    """Metaclass for FileLink deserialization."""

    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        cls.typ = properties.String('type', default="file_link", deserializable=False)


def _get_ids_from_url(url: str) -> Tuple[Optional[UUID], Optional[UUID]]:
    """Attempt to extract file_id and version_id from a URL."""
    parsed = urlparse(url)
    if len(parsed.query) > 0 or len(parsed.fragment) > 0:
        # Illegal modifiers
        return None, None
    split_path = urlparse(url).path.split('/')
    if len(split_path) >= 4 and split_path[-4] == 'files' and split_path[-2] == 'versions':
        file_id = split_path[-3]
        version_id = split_path[-1]
    elif len(split_path) >= 2 and split_path[-2] == 'files':
        file_id = split_path[-1]
        version_id = None
    else:
        file_id, version_id = None, None

    if file_id is not None:
        try:
            file_id = UUID(file_id)
            if version_id is not None:
                version_id = UUID(version_id)
        except ValueError:
            return None, None
    return file_id, version_id


class FileLink(
    GEMDResource['FileLink'],
    GEMDFileLink,
    metaclass=FileLinkMeta,
    typ=GEMDFileLink.typ
):
    """
    Resource that stores the name and url of an external file.

    Parameters
    ----------
    filename: str
        The name of the file.
    url: str
        URL that can be used to access the file.

    Attributes
    ----------
    uid: UUID
        Unique uuid4 identifier of this file; consistent across versions.
    version: UUID
        Unique uuid4 identifier of this version of this file
    version_number: Integer
        How many times this file has been uploaded;
        files are the "same" if the share a filename and dataset
    created_time: Datetime
        Time the file was created on platform.
    created_by: UUID
        Unique uuid4 identifier of this User who loaded this file
    mime_type: String
        Encoded string representing the type of the file (IETF RFC 2045)
    size: Integer
        Size in bytes of the file
    description: String
        A human-readable description of the file

    """

    # NOTE: skipping the "metadata" field since it appears to be unused
    # NOTE: skipping the "versioned_url" field since it is redundant
    # NOTE: skipping the "unversioned_url" field since it is redundant
    filename = properties.String('filename')
    url = properties.String('url')
    uid = properties.Optional(properties.UUID, 'id', serializable=False)
    version = properties.Optional(properties.UUID, 'version', serializable=False)
    created_time = properties.Optional(properties.Datetime, 'created_time', serializable=False)
    created_by = properties.Optional(properties.UUID, 'created_by', serializable=False)
    mime_type = properties.Optional(properties.String, 'mime_type', serializable=False)
    size = properties.Optional(properties.Integer, 'size', serializable=False)
    description = properties.Optional(properties.String, 'description', serializable=False)
    version_number = properties.Optional(properties.Integer, 'version_number', serializable=False)

    def __init__(self, filename: str, url: str):
        GEMDFileLink.__init__(self, filename, url)
        uid, version = _get_ids_from_url(url)
        if uid is not None:
            self.uid = uid
        if version is not None:
            self.version = version
        self.typ = FileLink.typ

    @classmethod
    def from_path(cls, path: Union[str, Path]) -> "FileLink":
        """Construct a FileLink from a local Path."""
        path = Path(path)  # In case it was a string
        return cls(filename=path.name, url=path.expanduser().absolute().as_uri())

    @property
    def name(self):
        """Attribute name is an alias for filename."""
        return self.filename

    @classmethod
    def _pre_build(cls, data: dict) -> dict:
        """Run data modification before building."""
        if 'url' in data and 'id' not in data:
            uid, version = _get_ids_from_url(data['url'])
            if uid is not None:
                data['id'] = str(uid)
            if version is not None:
                data['version'] = str(version)
        return data

    def __str__(self):
        return f'<File link {self.filename!r}>'


class FileCollection(Collection[FileLink]):
    """Represents the collection of all file links associated with a dataset."""

    _path_template = 'projects/{project_id}/datasets/{dataset_id}/files'
    _collection_key = 'files'
    _resource = FileLink

    def __init__(self, project_id: UUID, dataset_id: UUID, session: Session):
        self.project_id = project_id
        self.dataset_id = dataset_id
        self.session = session

    def _get_path(self,
                  uid: Optional[Union[UUID, str]] = None,
                  *,
                  ignore_dataset: Optional[bool] = False,
                  version: Union[str, UUID] = None,
                  action: Union[str, Sequence[str]] = [],
                  query_terms: Dict[str, str] = {},) -> str:
        """Build the path for taking an action with a particular file version."""
        if version is not None:
            action = ['versions', version] + ([action] if isinstance(action, str) else action)
        return super()._get_path(uid=uid, ignore_dataset=ignore_dataset, action=action)

    def _get_path_from_file_link(self, file_link: FileLink,
                                 *,
                                 action: str = None) -> str:
        """Build the platform path for taking an action with a particular file link."""
        if not self._is_on_platform_url(file_link.url) or file_link.uid is None:
            raise ValueError("FileLink did not contain a Citrine platform file URL.")
        return self._get_path(uid=file_link.uid, version=file_link.version, action=action)

    def build(self, data: dict) -> FileLink:
        """Build an instance of FileLink."""
        # Use this chance to construct a URL from platform metadata
        if 'url' not in data:
            data['url'] = self._get_path(uid=data['id'], version=data['version'])

        return FileLink.build(data)

    def get(self,
            uid: Union[UUID, str],
            *,
            version: Optional[Union[UUID, str, int]] = None) -> FileLink:
        """
        Retrieve an on-platform FileLink from its filename or file uuid.

        Parameters
        ----------
        uid: Union[UUID, str]
            A representation of the FileLink (Citrine id or file name)
        version: Optional[UUID, str, int]
            The version, as a UUID or str(UUID) of the version_id or an int or
            str(int) of the version number.  If None, returns the file with the
            highest version number (most recent).

        Returns
        -------
        ResourceType
            An object with specified scope and uid

        """
        if not isinstance(uid, (str, UUID)):
            raise TypeError(f"File Link can only be resolved from str or UUID."
                            f"Instead got {type(uid)} {uid}.")
        if version is not None and not isinstance(version, (str, UUID, int)):
            raise TypeError(f"Version can only be resolved from str, int or UUID."
                            f"Instead got {type(uid)} {uid}.")

        if isinstance(uid, str):
            try:  # Check if the uid string is actually a UUID
                uid = UUID(uid)
            except ValueError:
                pass

        if isinstance(version, str):
            try:  # Check if the version string is actually a UUID
                version = UUID(version)
            except ValueError:
                try:  # Check if the version string is actually an int / version number
                    version = int(version)
                except ValueError:
                    raise ValueError(
                        f"Version {version} could not be converted to either an int or a UUID"
                    )

        if isinstance(uid, str):
            # Assume it's the filename on platform;
            if version is None or isinstance(version, int):
                file = self._search_by_file_name(dset_id=self.dataset_id,
                                                 file_name=uid,
                                                 file_version_number=version)
            else:  # We did our type checks earlier; version is a UUID
                file = self._search_by_file_version_id(file_version_id=version)
        else:  # We did our type checks earlier; uid is a UUID
            if isinstance(version, UUID):
                file = self._search_by_file_version_id(file_version_id=version)
            else:  # We did our type checks earlier; version is an int or None
                file = self._search_by_dataset_file_id(dataset_file_id=uid,
                                                       dset_id=self.dataset_id,
                                                       file_version_number=version)

        return file

    def upload(self, *, file_path: Union[str, Path], dest_name: str = None) -> FileLink:
        """
        Uploads a file to the dataset.

        Parameters
        ----------
        file_path: str, Path
            The path to the file on the local computer.
        dest_name: str, optional
            The name the file will have after being uploaded. If unspecified, the local name of
            the file will be used. That is, the file at "/Users/me/diagram.pdf" will be uploaded
            with the name "diagram.pdf". File names **must be unique** within a dataset. If a file
            is uploaded with the same `dest_name` as an existing file it will be considered
            a new version of the existing file.

        Returns
        -------
        FileLink
            The filename and url of the uploaded object.

        """
        file_path = Path(file_path).expanduser()
        if not file_path.is_file():
            raise ValueError(f"{file_path} is not a file.")

        if not dest_name:
            # Use the file name as a default dest_name
            dest_name = file_path.name

        uploader = self._make_upload_request(file_path, dest_name)
        uploader = self._upload_file(file_path, uploader)
        return self._complete_upload(dest_name, uploader)

    def _make_upload_request(self, file_path: Path, dest_name: str):
        """
        Make a request to the backend to upload a file. Uses mimetypes.guess_type.

        Parameters
        ----------
        file_path: Path
            The path to the file on the local computer.
        dest_name: str
            The name the file will have after being uploaded.

        Returns
        -------
        _Uploader
            Holds the parameters returned by the upload request, for later use.
            These must include region_name, aws_access_key_id, aws_secret_access_key,
            aws_session_token, bucket, object_key, & upload_id.

        """
        path = self._get_path(action="uploads")
        mime_type = self._mime_type(file_path)
        file_size = file_path.stat().st_size
        assert isinstance(file_size, int)
        upload_json = {
            'files': [
                {
                    'file_name': dest_name,
                    'mime_type': mime_type,
                    'size': file_size
                }
            ]
        }
        # POST request creates space in S3 for the file and returns AWS-related information
        # (such as temporary credentials) that allow the file to be uploaded.
        upload_request = self.session.post_resource(path=path, json=upload_json)
        uploader = _Uploader()

        # Extract all relevant information from the upload request
        try:

            uploader.region_name = upload_request['s3_region']
            uploader.aws_access_key_id = upload_request['temporary_credentials']['access_key_id']
            uploader.aws_secret_access_key = \
                upload_request['temporary_credentials']['secret_access_key']
            uploader.aws_session_token = upload_request['temporary_credentials']['session_token']
            uploader.bucket = upload_request['s3_bucket']
            uploader.object_key = upload_request['uploads'][0]['s3_key']
            uploader.upload_id = upload_request['uploads'][0]['upload_id']
            uploader.s3_endpoint_url = self.session.s3_endpoint_url
            uploader.s3_use_ssl = self.session.s3_use_ssl
            uploader.s3_addressing_style = self.session.s3_addressing_style

        except KeyError:
            raise RuntimeError("Upload initiation response is missing some fields: "
                               "{}".format(upload_request))
        return uploader

    def _search_by_file_name(self,
                             file_name: str,
                             dset_id: UUID,
                             file_version_number: Optional[int] = None
                             ) -> Optional[FileLink]:
        """
        Make a request to the backend to search a file by name.

        Note that you can specify a version number, in case you don't, it will
        return the last version by default.

        Parameters
        ----------
        file_name: str
            The name of the file.
        dset_id: UUID
            UUID that represents a dataset.
        file_version_number: Optional[int]
            As optional, you can send a specific version number.

        Returns
        -------
        FileLink
            All the data needed for a file.

        """
        path = self._get_path(action="search")

        search_json = {
            'fileSearchFilter':
                {
                    'type': SearchFileFilterTypeEnum.NAME_SEARCH.value,
                    'datasetId': str(dset_id),
                    'fileName': file_name,
                    'fileVersionNumber': file_version_number
                }
        }

        data = self.session.post_resource(path=path, json=search_json)

        return self.build(data['files'][0])

    def _search_by_file_version_id(self,
                                   file_version_id: UUID
                                   ) -> Optional[FileLink]:
        """
        Make a request to the backend to search a file by file version id.

        Parameters
        ----------
        file_version_id: UUID
            UUID that represents a file version id.

        Returns
        -------
        FileLink
            All the data needed for a file.

        """
        path = self._get_path(action="search")

        search_json = {
            'fileSearchFilter': {
                'type': SearchFileFilterTypeEnum.VERSION_ID_SEARCH.value,
                'fileVersionUuid': str(file_version_id)
            }
        }

        data = self.session.post_resource(path=path, json=search_json)

        return self.build(data['files'][0])

    def _search_by_dataset_file_id(self,
                                   dataset_file_id: UUID,
                                   dset_id: UUID,
                                   file_version_number: Optional[int] = None
                                   ) -> Optional[FileLink]:
        """
        Make a request to the backend to search a file by dataset file id.

        Note that you can specify a version number, in case you don't, it will
        return the last version by default.

        Parameters
        ----------
        dataset_file_id: UUID
            UUID that represents a dataset file id.
        dset_id: UUID
            UUID that represents a dataset.
        file_version_number: Optional[int]
            As optional, you can send a specific version number

        Returns
        -------
        FileLink
            All the data needed for a file.

        """
        path = self._get_path(action="search")

        search_json = {
            'fileSearchFilter': {
                'type': SearchFileFilterTypeEnum.DATASET_FILE_ID_SEARCH.value,
                'datasetId': str(dset_id),
                'datasetFileId': str(dataset_file_id),
                'fileVersionNumber': file_version_number
            }
        }

        data = self.session.post_resource(path=path, json=search_json)

        return self.build(data['files'][0])

    @staticmethod
    def _mime_type(file_path: Path):
        # This string coercion is for supporting pathlib.Path objects in python < 3.8
        mime_type = mimetypes.guess_type(str(file_path))[0]
        if mime_type is None:
            mime_type = "application/octet-stream"
        return mime_type

    @staticmethod
    def _upload_file(file_path: Path, uploader: _Uploader):
        """
        Upload a file to S3.

        Parameters
        ----------
        file_path: Path
            The path to the file on the local computer.
        uploader: _Uploader
            Holds the parameters returned by the upload request.

        Returns
        -------
        _Uploader
            The input uploader object with its s3_version field now populated.

        """
        additional_s3_opts = {
            'use_ssl': uploader.s3_use_ssl,
            'config': Config(s3={'addressing_style': uploader.s3_addressing_style})
        }

        if uploader.s3_endpoint_url is not None:
            additional_s3_opts['endpoint_url'] = uploader.s3_endpoint_url

        s3_client = boto3_client('s3',
                                 region_name=uploader.region_name,
                                 aws_access_key_id=uploader.aws_access_key_id,
                                 aws_secret_access_key=uploader.aws_secret_access_key,
                                 aws_session_token=uploader.aws_session_token,
                                 **additional_s3_opts)
        with file_path.open(mode='rb') as f:
            try:
                # NOTE: This is only using the simple PUT logic, not the more sophisticated
                # multipart upload approach that is also available (providing parallel
                # uploads, etc).
                upload_response = s3_client.put_object(
                    Bucket=uploader.bucket,
                    Key=uploader.object_key,
                    Body=f,
                    Metadata={"X-Citrine-Upload-Id": uploader.upload_id})
            except ClientError as e:
                raise RuntimeError(f"Upload of file {file_path} failed with the following "
                                   f"exception: {e}")
        uploader.s3_version = upload_response['VersionId']
        return uploader

    def _complete_upload(self, dest_name: str, uploader: _Uploader):
        """
        Indicate that the upload has finished and determine the file URL.

        Parameters
        ----------
        dest_name: str
            The name the file will have after being uploaded.
        uploader: _Uploader
            Holds the parameters returned by the upload request and the upload response.

        Returns
        -------
        FileLink
            The filename and url of the uploaded object.

        """
        url = self._get_path(action=["uploads", uploader.upload_id, "complete"])
        complete_response = self.session.put_resource(path=url,
                                                      json={'s3_version': uploader.s3_version})

        try:
            file_id = complete_response['file_info']['file_id']
            version_id = complete_response['file_info']['version']
        except KeyError:
            raise RuntimeError("Upload completion response is missing some "
                               "fields: {}".format(complete_response))

        return self.build({"filename": dest_name, "id": file_id, "version": version_id})

    def download(self, *, file_link: Union[str, UUID, FileLink], local_path: Union[str, Path]):
        """
        Download the file associated with a given FileLink to the local computer.

        Parameters
        ----------
        file_link: FileLink, str, UUID
            Resource referencing the file.
        local_path: str, Path
            Path to save file on the local computer. If `local_path` is a directory,
            then the filename of this FileLink object will be appended to the path.

        """
        file_link = self._resolve_file_link(file_link)

        if isinstance(local_path, str):
            directory, filename = os.path.split(local_path)
            if len(filename) == 0:  # the string ended with /
                final_path = Path(directory) / file_link.filename
            else:
                final_path = Path(directory) / filename
        elif local_path.is_dir():
            final_path = local_path / file_link.filename
        else:
            final_path = local_path

        response_content = self.read(file_link=file_link)
        write_file_locally(response_content, final_path)

    def read(self, *, file_link: Union[str, UUID, FileLink]) -> bytes:
        """
        Read the file associated with a given FileLink.

        Parameters
        ----------
        file_link: FileLink, str, UUID
            Resource referencing the file.

        Returns
        -------
        I/O stream
            The contents of the file.

        """
        file_link = self._resolve_file_link(file_link)

        if self._is_local_url(file_link.url):  # Read the local file
            path = Path(unquote_plus(urlparse(file_link.url).path))
            return path.read_bytes()

        if self._is_external_url(file_link.url):  # Pull it from where ever it lives
            final_url = file_link.url
        else:
            # The "/content-link" route returns a pre-signed url to download the file.
            content_link = self._get_path_from_file_link(file_link, action='content-link')
            content_link_response = self.session.get_resource(content_link)
            pre_signed_url = content_link_response['pre_signed_read_link']
            final_url = rewrite_s3_links_locally(pre_signed_url, self.session.s3_endpoint_url)

        download_response = requests.get(final_url)
        return download_response.content

    @deprecated(deprecated_in="2.4.0",
                removed_in="3.0.0",
                details="The process file protocol is deprecated "
                        "in favor of ingest()")
    def process(self, *, file_link: Union[FileLink, str, UUID],
                processing_type: FileProcessingType,
                wait_for_response: bool = True,
                timeout: float = 2 * 60,
                polling_delay: float = 1.0) -> Union[JobSubmissionResponse,
                                                     Dict[FileProcessingType,
                                                          FileProcessingResult]]:
        """
        Start a File Processing async job, returning a pollable job response.

        Parameters
        ----------
        file_link: FileLink, str, or UUID
            The file to process.
        processing_type: FileProcessingType
            The type of file processing to invoke.
        wait_for_response: bool
            Whether to wait for a result, vs. just return a job handle.  Default: True
        timeout: float
            How long to poll for the result before giving up. This is expressed in
            (fractional) seconds.  Default: 120 seconds.
        polling_delay: float
            How long to delay between each polling retry attempt.  Default: 1 second.

        Returns
        -------
        JobSubmissionResponse
            A JobSubmissionResponse which can be used to poll for the result.

        """
        if self._is_external_url(file_link.url):
            raise ValueError(f"Only on-platform resources can be processed. "
                             f"Passed URL {file_link.url}.")
        file_link = self._resolve_file_link(file_link)

        params = {"processing_type": processing_type.value}
        response = self.session.put_resource(
            self._get_path_from_file_link(file_link, action="processed"),
            json={},
            params=params
        )
        job = JobSubmissionResponse.build(response)
        logger.info('Build job submitted with job ID {}.'.format(job.job_id))

        if wait_for_response:
            return self.poll_file_processing_job(file_link=file_link,
                                                 processing_type=processing_type,
                                                 job_id=job.job_id, timeout=timeout,
                                                 polling_delay=polling_delay)
        else:
            return job

    @deprecated(deprecated_in="2.4.0",
                removed_in="3.0.0",
                details="The process file protocol is deprecated "
                        "in favor of ingest()")
    def poll_file_processing_job(self, *, file_link: FileLink,
                                 processing_type: FileProcessingType,
                                 job_id: UUID,
                                 timeout: float = 2 * 60,
                                 polling_delay: float = 1.0) -> Dict[FileProcessingType,
                                                                     FileProcessingResult]:
        """
        Poll for the result of the file processing task.

        Parameters
        ----------
        file_link: FileLink
            The file to process.
        processing_type: FileProcessingType
            The processing algorithm to apply.
        job_id: UUID
           The background job ID to poll for.
        timeout:
            How long to poll for the result before giving up. This is expressed in
            (fractional) seconds.
        polling_delay:
            How long to delay between each polling retry attempt.

        Returns
        -------
        None
           This method will raise an appropriate exception if the job failed, else
           it will return None to indicate the job was successful.

        """
        # Poll for job completion - this will raise an error if the job failed
        _poll_for_job_completion(self.session, self.project_id, job_id, timeout=timeout,
                                 polling_delay=polling_delay)

        return self.file_processing_result(file_link=file_link, processing_types=[processing_type])

    @deprecated(deprecated_in="2.4.0",
                removed_in="3.0.0",
                details="The process file protocol is deprecated "
                        "in favor of ingest()")
    def file_processing_result(self, *,
                               file_link: FileLink,
                               processing_types: List[FileProcessingType]) -> \
            Dict[FileProcessingType, FileProcessingResult]:
        """
        Return the file processing result for the given file link and processing type.

        Parameters
        ----------
        file_link: FileLink
            The file to process
        processing_types: FileProcessingType
            A list of the particular file processing types to retrieve

        Returns
        -------
        Map[FileProcessingType, FileProcessingResult]
            The file processing results, mapped by processing type.

        """
        processed_results_path = self._get_path_from_file_link(file_link, action="processed")

        params = []
        for proc_type in processing_types:
            params.append(('processing_type', proc_type.value))

        response = self.session.get_resource(processed_results_path, params=params)
        results_json = response['results']
        results = {}
        for result_json in results_json:
            processing_type = FileProcessingType[result_json['processing_type']]
            data = result_json['data']

            if processing_type == FileProcessingType.VALIDATE_CSV:
                data = CsvValidationData.build(data)

            result = FileProcessingResult(processing_type, data)
            results[processing_type] = result

        return results

    def ingest(self,
               files: Iterable[Union[FileLink, Path, str]],
               *,
               upload: bool = False,
               raise_errors: bool = True,
               build_table: bool = False,
               delete_dataset_contents: bool = False,
               delete_templates: bool = True,
               timeout: float = None,
               polling_delay: Optional[float] = None
               ) -> "IngestionStatus":  # noqa: F821
        """
        [ALPHA] Ingest a set of CSVs and/or Excel Workbooks formatted per the gemd-ingest protocol.

        Parameters
        ----------
        files: List[FileLink, Path, or str]
            A list of files from which GEMD objects should be built.
            A FileLink must contain an absolute URL or a relative path for an on-platform resource.
            Strings must be resolvable to a FileLink; if resolution fails, an exception is thrown.
            * If upload is False, an attempt is made to resolve it to an on-platform resource.
            * If upload is True, resolves locally first, and falls back to on-platform.
        upload: bool
            If the files are off-platform references, upload them first.
            Defaults to False, in which case an off-platform resource raises an error.
            A file is off-platform if it has an absolute URL and that URL is not for the
            citrine platform.
            For example, https://example.com/file.csv is off-platform, and would be
            uploaded to platform and ingested if upload is True.
        raise_errors: bool
            Whether ingestion errors raise exceptions (vs. simply reported in the results).
            Default: True
        build_table: bool
            Whether to trigger a regeneration of the table config and building the table
            after ingestion.  Default: False
        delete_dataset_contents: bool
            Whether to delete old objects prior to creating new ones.  Default: False
        delete_templates: bool
            Whether to delete old templates if deleting old objects.  Default: True
        timeout: Optional[float]
            Amount of time to wait on the job (in seconds) before giving up. Note that
            this number has no effect on the underlying job itself, which can also time
            out server-side.
        polling_delay: Optional[float]
            How long to delay between each polling retry attempt.


        Returns
        ----------
        IngestionStatus
            The result of the ingestion operation.

        """
        from citrine.resources.ingestion import IngestionCollection

        def resolve_with_local(candidate: Union[FileLink, Path, str]) -> FileLink:
            """Resolve Path, str or FileLink to an absolute reference."""
            if upload:
                if not isinstance(candidate, GEMDFileLink):
                    try:
                        absolute = Path(candidate).expanduser().resolve(strict=True)
                        return FileLink(filename=absolute.name, url=absolute.as_uri())
                    except FileNotFoundError as e:
                        if isinstance(candidate, Path):  # Unresolvable path passed
                            raise e
            elif isinstance(candidate, Path):
                raise TypeError("Path objects are only valid when upload=True.")

            return self._resolve_file_link(candidate)  # It was a FileLink or unresolvable string

        resolved = []
        for file in files:
            resolved_file = resolve_with_local(file)
            if resolved_file not in resolved:  # Remove duplicates
                resolved.append(resolved_file)

        onplatform = [f for f in resolved if self._is_on_platform_url(f.url)]
        offplatform = [f for f in resolved if not self._is_on_platform_url(f.url)]

        if upload:
            with TemporaryDirectory() as downloads:
                for file_link in offplatform:
                    path = Path(downloads) / file_link.filename
                    self.download(file_link=file_link, local_path=path)
                    onplatform.append(self.upload(file_path=path, dest_name=file_link.filename))
        elif len(offplatform) > 0:
            raise ValueError(f"All files must be on-platform to load them.  "
                             f"The following are not: {offplatform}")

        ingestion_collection = IngestionCollection(project_id=self.project_id,
                                                   dataset_id=self.dataset_id,
                                                   session=self.session)
        ingestion = ingestion_collection.build_from_file_links(
            file_links=onplatform,
            raise_errors=raise_errors
        )
        return ingestion.build_objects(
            build_table=build_table,
            delete_dataset_contents=delete_dataset_contents,
            delete_templates=delete_templates,
            timeout=timeout,
            polling_delay=polling_delay
        )

    def delete(self, file_link: FileLink):
        """
        Delete the file associated with a given FileLink from the database.

        Parameters
        ----------
        file_link: FileLink
            Resource referencing the external file.

        """
        if not self._is_on_platform_url(file_link.url):
            raise ValueError(f"Only platform resources can be deleted; passed URL {file_link.url}")
        file_id = _get_ids_from_url(file_link.url)[0]
        if file_id is None:
            raise ValueError(f"URL was malformed for local resources; passed URL {file_link.url}")
        data = self.session.delete_resource(self._get_path(file_id))
        return Response(body=data)

    def _resolve_file_link(self, identifier: Union[str, UUID, FileLink]) -> FileLink:
        """Generate the FileLink object referenced by the passed argument."""
        if isinstance(identifier, GEMDFileLink):
            if not isinstance(identifier, FileLink):  # Up-convert type with existing info
                update = FileLink(filename=identifier.filename, url=identifier.url)

                if self._is_on_platform_url(update.url):
                    if update.uid is None:
                        raise ValueError(f"URL was malformed for platform resources; "
                                         f"passed URL {update.url}")
                    else:  # Validate that it's a real record
                        update = self.get(uid=update.uid, version=update.version)
                        if update.filename != identifier.filename:
                            raise ValueError(
                                f"Name mismatch between link ({identifier.filename}) "
                                f"and platform ({update.filename})"
                            )
                    identifier = update

            return identifier  # Passthrough since it's as full as it can get
        elif isinstance(identifier, str):
            if self._is_on_platform_url(identifier):  # It's either a filename or URL
                file_id, version_id = _get_ids_from_url(identifier)
                if file_id is None:  # Try it as a name or a stand-alone UID
                    return self.get(identifier)
                else:  # We got a file UID (and possibly a version UID) from a URL
                    return self.get(uid=file_id, version=version_id)
            else:  # Assume it's an absolute URL
                filename = urlparse(identifier).path.split('/')[-1]
                return FileLink(filename=filename, url=identifier)
        elif isinstance(identifier, UUID):  # File UID
            return self.get(uid=identifier)
        else:
            raise TypeError(f"File Link can only be resolved from str, or UUID."
                            f"Instead got {type(identifier)} {identifier}.")

    def _is_external_url(self, url: str):
        """Check if the URL is absolute and not associated with this platform instance."""
        parsed = urlparse(url)
        if len(parsed.scheme) == 0 or len(parsed.netloc) == 0:
            # Relative
            return False

        return urlparse(self._get_path()).netloc != parsed.netloc

    @staticmethod
    def _is_local_url(url: str):
        """Check if the URL is for a local file."""
        parsed = urlparse(url)
        return parsed.scheme == "file"

    def _is_on_platform_url(self, url: str):
        """Check if the URL is for an on-platform record."""
        return not self._is_external_url(url) and not self._is_local_url(url)



from functools import partial
from typing import Optional, Iterable
from uuid import UUID

from citrine.informatics.executions.execution import Execution
from citrine.informatics.design_candidate import SampleSearchSpaceResultCandidate
from citrine._rest.resource import Resource
from citrine._utils.functions import format_escaped_url


class SampleDesignSpaceExecution(Resource['SampleDesignSpaceExecution'], Execution):
    """The execution of a Sample Design Space task.

    Possible statuses are INPROGRESS, SUCCEEDED, and FAILED.
    Additional fields ``status_description`` and ``status_detail`` provide more information.

    """

    design_space_id: Optional[UUID] = None

    def _path(self):
        return format_escaped_url(
            '/projects/{project_id}/design-spaces/{design_space_id}/sample/',
            project_id=self.project_id,
            design_space_id=self.design_space_id,
        )

    @classmethod
    def _pre_build(cls, data: dict) -> dict:
        """Run data modification before building."""
        # Flatten the status object in order to match other workflow objects.
        return {
            **data,
            "status_description": data['status']['minor'],
            "status_detail": data['status']['detail'],
            "status": data['status']["major"]
        }

    @classmethod
    def _build_results(
        cls, subset_collection: Iterable[dict]
    ) -> Iterable[SampleSearchSpaceResultCandidate]:
        for sample_result in subset_collection:
            yield SampleSearchSpaceResultCandidate.build(sample_result)

    def results(
        self,
        *,
        page: Optional[int] = None,
        per_page: int = 100,
    ) -> Iterable[SampleSearchSpaceResultCandidate]:
        """Fetch the Sample Design Space Results for the particular execution, paginated."""
        path = self._path() + f'{self.uid}/results'
        fetcher = partial(self._fetch_page, path=path, fetch_func=self._session.get_resource)
        return self._paginator.paginate(page_fetcher=fetcher,
                                        collection_builder=self._build_results,
                                        per_page=per_page)

    def result(
        self,
        *,
        result_id: UUID,
    ) -> SampleSearchSpaceResultCandidate:
        """Fetch a Sample Design Space Result for the particular UID."""
        path = self._path() + f'{self.uid}/results/{result_id}'
        data = self._session.get_resource(path, version=self._api_version)
        result = SampleSearchSpaceResultCandidate.build(data)
        return result



from logging import getLogger
from time import time, sleep
from typing import Union
from uuid import UUID

from citrine._serialization.properties import Set as PropertySet, String, Object
from citrine._rest.resource import Resource
from citrine._serialization import properties
from citrine._session import Session
from citrine._utils.functions import format_escaped_url
from citrine.exceptions import PollingTimeoutError, JobFailureError

logger = getLogger(__name__)


class JobSubmissionResponse(Resource['JobSubmissionResponse']):
    """A response to a submit-job request for the job submission framework.

    This is returned as a successful response from the remote service.
    """

    job_id = properties.UUID("job_id")
    """:UUID: job id of the job submission request"""


class TaskNode(Resource['TaskNode']):
    """Individual task status.

    The TaskNode describes a component of an overall job.
    """

    id = properties.String("id")
    """:str: unique identification number for the job task"""
    task_type = properties.String("task_type")
    """:str: the type of task running"""
    status = properties.String("status")
    """:str: The last reported status of this particular task.
    One of "Submitted", "Pending", "Running", "Success", or "Failure"."""
    dependencies = PropertySet(String(), "dependencies")
    """:Set[str]: all the tasks that this task is dependent on"""
    failure_reason = properties.Optional(String(), "failure_reason")
    """:str: if a task has failed, the failure reason will be in this parameter"""


class JobStatusResponse(Resource['JobStatusResponse']):
    """A response to a job status check.

    The JobStatusResponse summarizes the status for the entire job.
    """

    job_type = properties.String("job_type")
    """:str: the type of job for this status report"""
    status = properties.String("status")
    """:str: The status of the job. One of "Running", "Success", or "Failure"."""
    tasks = properties.List(Object(TaskNode), "tasks")
    """:List[TaskNode]: all of the constituent task required to complete this job"""
    output = properties.Optional(properties.Mapping(String, String), 'output')
    """:Optional[dict[str, str]]: job output properties and results"""


def _poll_for_job_completion(session: Session,
                             project_id: Union[UUID, str],
                             job: Union[JobSubmissionResponse, UUID, str],
                             *,
                             timeout: float = 2 * 60,
                             polling_delay: float = 2.0,
                             raise_errors: bool = True,
                             ) -> JobStatusResponse:
    """
    Polls for job completion given a timeout.

    This polls for job completion given the Job ID, failing appropriately if the job result
    was not successful.

    Parameters
    ----------
    job
        The job submission object or job ID that was given from a job submission.
    timeout
        Amount of time to wait on the job (in seconds) before giving up. Defaults
        to 2 minutes. Note that this number has no effect on the underlying job
        itself, which can also time out server-side.
    polling_delay:
        How long to delay between each polling retry attempt.
    raise_errors:
        Whether a `Failure` response should raise a JobFailureError.

    Returns
    -------
    JobStatusResponse
        The job response information that can be used to extract relevant
        information from the completed job.

    """
    if isinstance(job, JobSubmissionResponse):
        job_id = job.job_id
    else:
        job_id = job  # pragma: no cover
    path = format_escaped_url('projects/{}/execution/job-status', project_id)
    params = {'job_id': job_id}
    start_time = time()
    while True:
        response = session.get_resource(path=path, params=params)
        status: JobStatusResponse = JobStatusResponse.build(response)
        if status.status in ['Success', 'Failure']:
            break
        elif time() - start_time < timeout:
            logger.info(
                f'Job still in progress, polling status again in {polling_delay:.2f} seconds.'
            )

            sleep(polling_delay)
        else:
            logger.error(f'Job exceeded user timeout of {timeout} seconds. '
                         f'Note job on server is unaffected by this timeout.')
            logger.debug('Last status: {}'.format(status.dump()))
            raise PollingTimeoutError('Job {} timed out.'.format(job_id))
    if status.status == 'Failure':
        logger.debug(f'Job terminated with Failure status: {status.dump()}')
        if raise_errors:
            failure_reasons = []
            for task in status.tasks:
                if task.status == 'Failure':
                    logger.error(f'Task {task.id} failed with reason "{task.failure_reason}"')
                    failure_reasons.append(task.failure_reason)
            raise JobFailureError(
                message=f'Job {job_id} terminated with Failure status. '
                        f'Failure reasons: {failure_reasons}',
                job_id=job_id,
                failure_reasons=failure_reasons)

    return status



"""Column definitions for GEM Tables."""
from typing import Type, Optional, List, Union
from abc import abstractmethod

from gemd.enumeration.base_enumeration import BaseEnumeration

from citrine._serialization.serializable import Serializable
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._serialization import properties
from citrine.gemtables.variables import Variable


class CompositionSortOrder(BaseEnumeration):
    """Order to use when sorting the components in a composition.

    * ``ALPHABETICAL`` is alpha-numeric order by the component name
    * ``QUANTITY`` is ordered from the largest to smallest quantity, with ties
      broken alphabetically
    """

    ALPHABETICAL = "alphabetical"
    QUANTITY = "quantity"


class ChemicalDisplayFormat(BaseEnumeration):
    """Format to use when rendering a molecular structure.

    * ``SMILES`` Simplified molecular-input line-entry system
    * ``INCHI`` International Chemical Identifier
    """

    SMILES = "smiles"
    INCHI = "inchi"


def _make_data_source(variable_rep: Union[str, Variable]) -> str:
    """Return a string appropriate to use as a data_source.

    Parameters
    ----------
    variable_rep: Union[str, Variable]
        Either the name of the variable or the variable itself

    """
    if isinstance(variable_rep, str):
        return variable_rep
    elif isinstance(variable_rep, Variable):
        return variable_rep.name
    else:
        raise TypeError("Columns can only be linked by str or Variable."
                        "Instead got {}.".format(variable_rep))


class Column(PolymorphicSerializable['Column']):
    """A column in the GEM Table, defined as some operation on a variable.

    Abstract type that returns the proper type given a serialized dict.
    """

    @abstractmethod
    def _attrs(self) -> List[str]:
        pass  # pragma: no cover

    def __eq__(self, other):
        try:
            return all([
                self.__getattribute__(key) == other.__getattribute__(key) for key in self._attrs()
            ])
        except AttributeError:
            return False

    @classmethod
    def get_type(cls, data) -> Type[Serializable]:
        """Return the subtype."""
        if "type" not in data:
            raise ValueError("Can only get types from dicts with a 'type' key")
        types: List[Type[Serializable]] = [
            IdentityColumn,
            MeanColumn, StdDevColumn, QuantileColumn, OriginalUnitsColumn,
            MostLikelyCategoryColumn, MostLikelyProbabilityColumn,
            FlatCompositionColumn, ComponentQuantityColumn,
            NthBiggestComponentNameColumn, NthBiggestComponentQuantityColumn,
            MolecularStructureColumn, ConcatColumn
        ]
        res = next((x for x in types if x.typ == data["type"]), None)
        if res is None:
            raise ValueError("Unrecognized type: {}".format(data["type"]))
        return res


class MeanColumn(Serializable['MeanColumn'], Column):
    """Column containing the mean of a real-valued variable.

    Parameters
    ----------
    data_source: Union[str, Variable]
        name of the variable to use when populating the column
    target_units: Optional[str]
        units to convert the real variable into. If not specified:
            a) If there is an OriginalUnitsColumnDefinition for that source,
                no conversion will be made.
            b) If not, the real variable will be converted by using the
                default_units from the associated template.

    """

    data_source = properties.String('data_source')
    target_units = properties.Optional(properties.String, "target_units")
    typ = properties.String('type', default="mean_column", deserializable=False)

    def _attrs(self) -> List[str]:
        return ["data_source", "target_units", "typ"]

    def __init__(self, *,
                 data_source: Union[str, Variable],
                 target_units: Optional[str] = None):
        self.data_source = _make_data_source(data_source)
        self.target_units = target_units


class StdDevColumn(Serializable["StdDevColumn"], Column):
    """Column containing the standard deviation of a real-valued variable.

    Parameters
    ----------
    data_source: Union[str, Variable]
        name of the variable to use when populating the column
    target_units: Optional[str]
        units to convert the real variable into. If not specified:
            a) If there is an OriginalUnitsColumnDefinition for that source,
                no conversion will be made.
            b) If not, the real variable will be converted by using the
                default_units from the associated template.

    """

    data_source = properties.String('data_source')
    target_units = properties.Optional(properties.String, "target_units")
    typ = properties.String('type', default="std_dev_column", deserializable=False)

    def _attrs(self) -> List[str]:
        return ["data_source", "target_units", "typ"]

    def __init__(self, *,
                 data_source: Union[str, Variable],
                 target_units: Optional[str] = None):
        self.data_source = _make_data_source(data_source)
        self.target_units = target_units


class QuantileColumn(Serializable["QuantileColumn"], Column):
    """Column containing a quantile of the variable.

    The column is populated with the quantile function of the distribution evaluated at "quantile".
    For example, for a uniform distribution parameterized by a lower and upper bound, the value
    in the column would be:

    .. math::

        lower + (upper - lower) * quantile

    while for a normal distribution parameterized by a mean and stddev, the value would be:

    .. math::

        mean + stddev * \\sqrt{2} * erf^{-1}(2 * quantile - 1)

    Parameters
    ----------
    data_source: Union[str, Variable]
        name of the variable to use when populating the column
    quantile: float
        the quantile to use for the column, defined between 0.0 and 1.0
    target_units: Optional[str]
        units to convert the real variable into. If not specified:
            a) If there is an OriginalUnitsColumnDefinition for that source,
                no conversion will be made.
            b) If not, the real variable will be converted by using the
                default_units from the associated template.

    """

    data_source = properties.String('data_source')
    quantile = properties.Float("quantile")
    target_units = properties.Optional(properties.String, "target_units")
    typ = properties.String('type', default="quantile_column", deserializable=False)

    def _attrs(self) -> List[str]:
        return ["data_source", "quantile", "target_units", "typ"]

    def __init__(self, *,
                 data_source: Union[str, Variable],
                 quantile: float,
                 target_units: Optional[str] = None):
        self.data_source = _make_data_source(data_source)
        self.quantile = quantile
        self.target_units = target_units


class OriginalUnitsColumn(Serializable["OriginalUnitsColumn"], Column):
    """Column containing the units as entered in the source data.

    Parameters
    ----------
    data_source: Union[str, Variable]
        name of the variable to use when populating the column

    """

    data_source = properties.String('data_source')
    typ = properties.String('type', default="original_units_column", deserializable=False)

    def _attrs(self) -> List[str]:
        return ["data_source", "typ"]

    def __init__(self, *, data_source: Union[str, Variable]):
        self.data_source = _make_data_source(data_source)


class MostLikelyCategoryColumn(Serializable["MostLikelyCategoryColumn"], Column):
    """Column containing the most likely category.

    Parameters
    ----------
    data_source: Union[str, Variable]
        name of the variable to use when populating the column

    """

    data_source = properties.String('data_source')
    typ = properties.String('type', default="most_likely_category_column", deserializable=False)

    def _attrs(self) -> List[str]:
        return ["data_source", "typ"]

    def __init__(self, *, data_source: Union[str, Variable]):
        self.data_source = _make_data_source(data_source)


class MostLikelyProbabilityColumn(Serializable["MostLikelyProbabilityColumn"], Column):
    """Column containing the probability of the most likely category.

    Parameters
    ----------
    data_source: Union[str, Variable]
        name of the variable to use when populating the column

    """

    data_source = properties.String('data_source')
    typ = properties.String('type', default="most_likely_probability_column", deserializable=False)

    def _attrs(self) -> List[str]:
        return ["data_source", "typ"]

    def __init__(self, *, data_source: Union[str, Variable]):
        self.data_source = _make_data_source(data_source)


class FlatCompositionColumn(Serializable["FlatCompositionColumn"], Column):
    """Column that flattens the composition into a string of names and quantities.

    The numeric formatting tries to be human readable. For example, if all of the quantities
    are round numbers like ``{"spam": 4.0, "eggs": 1.0}`` then the result omit the decimal points
    like ``"(spam)4(eggs)1"`` (if sort_order is by quantity).

    Parameters
    ----------
    data_source: Union[str, Variable]
        name of the variable to use when populating the column
    sort_order: CompositionSortOrder
        order with which to sort the components when generating the flat string

    """

    data_source = properties.String('data_source')
    sort_order = properties.Enumeration(CompositionSortOrder, 'sort_order')
    typ = properties.String('type', default="flat_composition_column", deserializable=False)

    def _attrs(self) -> List[str]:
        return ["data_source", "sort_order", "typ"]

    def __init__(self, *,
                 data_source: Union[str, Variable],
                 sort_order: CompositionSortOrder):
        self.data_source = _make_data_source(data_source)
        self.sort_order = sort_order


class ComponentQuantityColumn(Serializable["ComponentQuantityColumn"], Column):
    """Column that extracts the quantity of a given component.

    If the component is not present in the composition, then the value in the column will be 0.0.

    Parameters
    ----------
    data_source: Union[str, Variable]
        name of the variable to use when populating the column
    component_name: str
        name of the component from which to extract the quantity
    normalize: bool
        whether to normalize the quantity by the sum of all component amounts. Default is false

    """

    data_source = properties.String('data_source')
    component_name = properties.String("component_name")
    normalize = properties.Boolean("normalize")
    typ = properties.String('type', default="component_quantity_column", deserializable=False)

    def _attrs(self) -> List[str]:
        return ["data_source", "component_name", "normalize", "typ"]

    def __init__(self, *,
                 data_source: Union[str, Variable],
                 component_name: str,
                 normalize: bool = False):
        self.data_source = _make_data_source(data_source)
        self.component_name = component_name
        self.normalize = normalize


class NthBiggestComponentNameColumn(Serializable["NthBiggestComponentNameColumn"], Column):
    """Name of the Nth biggest component.

    If there are fewer than N components in the composition, then this column will be empty.

    Parameters
    ----------
    data_source: Union[str, Variable]
        name of the variable to use when populating the column
    n: int
        index of the component name to extract, starting with 1 for the biggest

    """

    data_source = properties.String('data_source')
    n = properties.Integer("n")
    typ = properties.String('type', default="biggest_component_name_column", deserializable=False)

    def _attrs(self) -> List[str]:
        return ["data_source", "n", "typ"]

    def __init__(self, *,
                 data_source: Union[str, Variable],
                 n: int):
        self.data_source = _make_data_source(data_source)
        self.n = n


class NthBiggestComponentQuantityColumn(Serializable["NthBiggestComponentQuantityColumn"], Column):
    """Quantity of the Nth biggest component.

    If there are fewer than N components in the composition, then this column will be empty.

    Parameters
    ----------
    data_source: Union[str, Variable]
        name of the variable to use when populating the column
    n: int
        index of the component quantity to extract, starting with 1 for the biggest
    normalize: bool
        whether to normalize the quantity by the sum of all component amounts. Default is false

    """

    data_source = properties.String('data_source')
    n = properties.Integer("n")
    normalize = properties.Boolean("normalize")
    typ = properties.String('type',
                            default="biggest_component_quantity_column", deserializable=False)

    def _attrs(self) -> List[str]:
        return ["data_source", "n", "normalize", "typ"]

    def __init__(self, *,
                 data_source: Union[str, Variable],
                 n: int,
                 normalize: bool = False):
        self.data_source = _make_data_source(data_source)
        self.n = n
        self.normalize = normalize


class IdentityColumn(Serializable['IdentityColumn'], Column):
    """Column containing the value of a string-valued variable.

    Parameters
    ----------
    data_source: Union[str, Variable]
        name of the variable to use when populating the column

    """

    data_source = properties.String('data_source')
    typ = properties.String('type', default="identity_column", deserializable=False)

    def _attrs(self) -> List[str]:
        return ["data_source", "typ"]

    def __init__(self, *, data_source: Union[str, Variable]):
        self.data_source = _make_data_source(data_source)


class MolecularStructureColumn(Serializable['MolecularStructureColumn'], Column):
    """Column containing a representation of a molecular structure.

    Parameters
    ----------
    data_source: Union[str, Variable]
        name of the variable to use when populating the column
    format: ChemicalDisplayFormat
        the format in which to display the molecular structure

    """

    data_source = properties.String('data_source')
    format = properties.Enumeration(ChemicalDisplayFormat, 'format')
    typ = properties.String('type', default="molecular_structure_column", deserializable=False)

    def _attrs(self) -> List[str]:
        return ["data_source", "format", "typ"]

    def __init__(self, *, data_source: Union[str, Variable], format: ChemicalDisplayFormat):
        self.data_source = _make_data_source(data_source)
        self.format = format


class ConcatColumn(Serializable['ConcatColumn'], Column):
    """Column that concatenates multiple values produced by a list- or set-valued variable.

    The input subcolumn need not exist elsewhere in the table config, and its parameters have
    no bearing on how the table is constructed. Only the type of column is relevant. That a
    complete Column object is required is simply a limitation of the current API.

    Parameters
    ----------
    data_source: Union[str, Variable]
        name of the variable to use when populating the column
    subcolumn: Column
        a column of the type of the individual values to be concatenated

    """

    data_source = properties.String('data_source')
    subcolumn = properties.Object(Column, 'subcolumn')
    typ = properties.String('type', default="concat_column", deserializable=False)

    def _attrs(self) -> List[str]:
        return ["data_source", "typ"]

    def __init__(self, *, data_source: Union[str, Variable], subcolumn: Column):
        self.data_source = _make_data_source(data_source)
        self.subcolumn = subcolumn



from os import environ
from typing import Optional

from citrine._session import Session
from citrine.resources.project import ProjectCollection
from citrine.resources.team import TeamCollection
from citrine.resources.user import UserCollection


class Citrine:
    """The entry point for interacting with the Citrine Platform.

    Parameters
    ----------
    api_key: str
        Unique key that allows a user to access the Citrine Platform.
        Default: environ.get('CITRINE_API_KEY')
    scheme: str
        Networking protocol; usually https.  Default: https
    host: str
        Host URL, generally '<your_site>.citrine-platform.com'.
        Default: environ.get('CITRINE_API_HOST')
    port: Optional[str]
        Optional networking port.  Default: None

    """

    def __init__(self,
                 api_key: str = None,
                 *,
                 scheme: str = None,
                 host: str = None,
                 port: Optional[str] = None):
        if api_key is None:
            api_key = environ.get('CITRINE_API_KEY')
        if scheme is None:
            scheme = 'https'

        if host is None:
            host = environ.get('CITRINE_API_HOST')
            if host is None:
                raise ValueError("No host passed and environmental "
                                 "variable CITRINE_API_HOST not set.")

        self.session: Session = Session(refresh_token=api_key,
                                        scheme=scheme,
                                        host=host,
                                        port=port
                                        )

    @property
    def projects(self) -> ProjectCollection:
        """Return a resource representing all visible projects."""
        return ProjectCollection(self.session)

    @property
    def users(self) -> UserCollection:
        """Return the collection of all users."""
        return UserCollection(self.session)

    @property
    def teams(self) -> TeamCollection:
        """Returns a resource representing all visible teams."""
        return TeamCollection(self.session)



"""Row definitions for GEM Tables."""
from typing import Type, List, Set, Union
from abc import abstractmethod
from uuid import UUID

from gemd.entity.link_by_uid import LinkByUID
from gemd.entity.template import MaterialTemplate

from citrine._serialization.serializable import Serializable
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._serialization import properties

from citrine.resources.data_concepts import _make_link_by_uid


class Row(PolymorphicSerializable['Row']):
    """A rule for defining rows in a GEM Table.

    Abstract type that returns the proper type given a serialized dict.
    """

    @abstractmethod
    def _attrs(self) -> List[str]:
        pass  # pragma: no cover

    def __eq__(self, other):
        try:
            return all([
                self.__getattribute__(key) == other.__getattribute__(key) for key in self._attrs()
            ])
        except AttributeError:
            return False

    @classmethod
    def get_type(cls, data) -> Type[Serializable]:
        """Return the subtype."""
        if "type" not in data:
            raise ValueError("Can only get types from dicts with a 'type' key")
        types: List[Type[Serializable]] = [
            MaterialRunByTemplate
        ]
        res = next((x for x in types if x.typ == data["type"]), None)
        if res is None:
            raise ValueError("Unrecognized type: {}".format(data["type"]))
        return res


class MaterialRunByTemplate(Serializable['MaterialRunByTemplate'], Row):
    """Rows corresponding to MaterialRuns, marked by their template.

    Parameters
    ----------
    templates: list[Union[UUID, str, LinkByUID, MaterialTemplate]]
        templates of materials to include
    tags: Set[str]
        optional list of tags for filtering. If a terminal material doesn't
        contain any of the tags it will be filtered out.

    """

    templates = properties.List(properties.Object(LinkByUID), "templates")
    typ = properties.String('type', default="material_run_by_template", deserializable=False)
    tags = properties.Optional(properties.Set(properties.String), "tags")

    template_type = Union[UUID, str, LinkByUID, MaterialTemplate]

    def _attrs(self) -> List[str]:
        return ["templates", "typ", "tags"]

    def __init__(self, *,
                 templates: List[template_type],
                 tags: Set[str] = None):

        self.templates = [_make_link_by_uid(x) for x in templates]
        self.tags = tags



"""Citrine-specific exceptions."""
from types import SimpleNamespace
from typing import Optional, List
from urllib.parse import urlencode
from uuid import UUID

from requests import Response


class CitrineException(Exception):
    """The base exception class for Citrine-Python exceptions."""

    pass


class NonRetryableException(CitrineException):
    """Indicates that a non-retryable error occurred."""

    pass


class RetryableException(CitrineException):
    """Indicates an error occurred but it is retryable."""

    pass


class UnauthorizedRefreshToken(NonRetryableException):
    """The token used to refresh authentication is invalid."""

    pass


class NonRetryableHttpException(NonRetryableException):
    """An exception originating from an HTTP error from a Citrine API."""

    def __init__(self, path: str, response: Optional[Response] = None):
        self.url = path
        self.detailed_error_info = []
        if response is not None:
            self.response_text = response.text
            self.code = response.status_code

            method = "unknown"
            if response.request is not None:
                method = response.request.method

            self.detailed_error_info.append(
                "{} (code: {}) returned from {} request to path: '{}'".format(
                    response.reason, self.code, method, path
                )
            )
            try:
                resp_json = response.json()
                if isinstance(resp_json, dict):
                    from citrine.resources.api_error import ApiError
                    self.api_error = ApiError.build(resp_json)

                    validation_error_msgs = [
                        "{} ({})".format(f.failure_message, f.failure_id)
                        for f in self.api_error.validation_errors]

                    if self.api_error.message is not None:
                        self.detailed_error_info.append(self.api_error.message)
                    self.detailed_error_info.extend(validation_error_msgs)
                else:
                    self.api_error = None
                    if response.text is not None:
                        self.detailed_error_info.append(response.text)
            except (TypeError, ValueError):
                # If the response does not have a json dictionary, use its text
                self.api_error = None
                if response.text is not None:
                    self.detailed_error_info.append(response.text)
        else:
            self.http_err = path
            self.response_text = None
            self.code = None
            self.api_error = None

        super().__init__("\n\t".join(self.detailed_error_info))


class NotFound(NonRetryableHttpException):
    """A particular url was not found. (http status 404)."""

    @staticmethod
    def build(*, message: str, method: str, path: str, params: dict = {}):
        """
        Construct a NotFound response from parameters.

        This allows raising NotFound locally to produce consistent behavior even if the server
        doesn't return a 404.

        Parameters
        ----------
        message: str
            The text of the exception message, which will be surrounded by more details.

        method: str
            The HTTP method which resulted in the error, to display in the full message.

        path: str
            The path from which the error came, to display in the full message

        params: dict
            The query parameters passed as part of the request, to display in the full message.

        Returns
        -------
        NotFound
            The NotFound exception corresponding to the above arguments.

        """
        path_and_query = path + (f"?{urlencode(params)}" if params else "")
        return NotFound(
            path_and_query,
            SimpleNamespace(
                text=message,
                status_code=404,
                request=SimpleNamespace(method=method.upper()),
                reason="Not Found",
                json=lambda self: {"code": 404, "message": message, "validation_errors": []}
            )
        )


class Unauthorized(NonRetryableHttpException):
    """The user is unauthorized to make this api call. (http status 401)."""

    pass


class BadRequest(NonRetryableHttpException):
    """The user is trying to perform an invalid operation. (http status 400)."""

    pass


class WorkflowConflictException(NonRetryableHttpException):
    """There is a conflict preventing the workflow from being executed. (http status 409)."""

    pass


# A 409 is a Conflict, and can be raised anywhere a conflict occurs, not just in a workflow.
Conflict = WorkflowConflictException


class WorkflowNotReadyException(RetryableException):
    """The workflow is not ready to be executed. I.e., still validating. (http status 425)."""

    pass


class PollingTimeoutError(NonRetryableException):
    """Polling for an asynchronous result has exceeded the timeout."""

    pass


class JobFailureError(NonRetryableException):
    """The asynchronous job completed with the given failure message."""

    def __init__(self, *, message: str, job_id: UUID, failure_reasons: List[str]):
        super().__init__(message)
        self.job_id = job_id
        self.failure_reasons = failure_reasons


class ModuleRegistrationFailedException(NonRetryableException):
    """A module failed to register."""

    def __init__(self, moduleType: str, exc: Exception):
        err = 'The "{0}" failed to register. {1}: {2}'.format(
            moduleType, exc.__class__.__name__, str(exc))
        super().__init__(err)



import platform
from datetime import datetime

import jwt
import pytest
import pytz
import requests_mock

from citrine import Citrine


def refresh_token(expiration: datetime = None) -> dict:
    token = jwt.encode(
        payload={'exp': expiration.timestamp()},
        key='garbage'
    )
    return {'access_token': token}


token_refresh_response = refresh_token(datetime(2019, 3, 14, tzinfo=pytz.utc))


def test_citrine_creation():
    with requests_mock.Mocker() as m:
        m.post('https://citrine-testing.fake/api/v1/tokens/refresh', json=token_refresh_response)

        assert '1234' == Citrine(api_key='1234', host='citrine-testing.fake').session.refresh_token


def test_citrine_signature(monkeypatch):
    with requests_mock.Mocker() as m:
        m.post('http://citrine-testing.fake:8080/api/v1/tokens/refresh', json=token_refresh_response)

        assert '1234' == Citrine(api_key='1234',
                                 scheme='http',
                                 host='citrine-testing.fake',
                                 port="8080").session.refresh_token

    # Validate defaults
    with requests_mock.Mocker() as m:
        patched_key = "5678"
        patched_host = "monkeypatch.citrine-testing.fake"
        monkeypatch.setenv("CITRINE_API_KEY", patched_key)
        monkeypatch.setenv("CITRINE_API_HOST", patched_host)
        m.post(f'https://{patched_host}/api/v1/tokens/refresh', json=token_refresh_response)

        assert patched_key == Citrine().session.refresh_token
        assert patched_key == Citrine(api_key=patched_key).session.refresh_token
        monkeypatch.delenv("CITRINE_API_KEY")
        assert Citrine().session.refresh_token is None

    monkeypatch.delenv("CITRINE_API_HOST")
    with pytest.raises(ValueError):
        Citrine()


def test_citrine_project_session():
    with requests_mock.Mocker() as m:
        m.post('https://citrine-testing.fake/api/v1/tokens/refresh', json=token_refresh_response)

        citrine = Citrine(api_key='foo', host='citrine-testing.fake')

    assert citrine.session == citrine.projects.session


def test_citrine_user_session():
    with requests_mock.Mocker() as m:
        m.post('https://citrine-testing.fake/api/v1/tokens/refresh', json=token_refresh_response)
        citrine = Citrine(api_key='foo', host='citrine-testing.fake')
    assert citrine.session == citrine.users.session


def test_citrine_team_session():
    with requests_mock.Mocker() as m:
        m.post('https://citrine-testing.fake/api/v1/tokens/refresh', json=token_refresh_response)
        citrine = Citrine(api_key='foo', host='citrine-testing.fake')
    assert citrine.session == citrine.teams.session


def test_citrine_user_agent():
    with requests_mock.Mocker() as m:
        m.post('https://citrine-testing.fake/api/v1/tokens/refresh', json=token_refresh_response)
        citrine = Citrine(api_key='foo', host='citrine-testing.fake')

    agent_parts = citrine.session.headers['User-Agent'].split()
    python_impls = {'CPython', 'IronPython', 'Jython', 'PyPy'}
    expected_products = {'python-requests', 'citrine-python'}

    for product in agent_parts:
        product_name, product_version = product.split('/')
        assert product_name in {*python_impls, *expected_products}

        if product_name in python_impls:
            assert product_version == platform.python_version()
        else:
            # Check that the version is major.minor.patch but don't
            # enforce them to be ints.  It's common to see strings used
            # as the patch version
            assert len(product_version.split('.')) == 3



from typing import List, Optional, Set

from gemd.enumeration.base_enumeration import BaseEnumeration

from citrine._rest.resource import Resource
from citrine._serialization import properties as _properties
from citrine.informatics.data_sources import DataSource
from citrine.informatics.descriptors import Descriptor
from citrine.informatics.predictors import PredictorNode
from citrine.informatics.predictors.node import _check_deprecated_training_data

__all__ = ['AutoMLPredictor', 'AutoMLEstimator']


class AutoMLEstimator(BaseEnumeration):
    """[ALPHA] Algorithms to be used during AutoML model selection.

    * LINEAR corresponds to a linear regression estimator
        (valid for single-task regression problems)
    * RANDOM_FOREST corresponds to a random forest estimator
        (valid for single-task and multi-task regression and classification problems)
    * GAUSSIAN_PROCESS corresponds to a Gaussian process estimator
        (valid for single-task regression and classification problems)
    * SUPPORT_VECTOR_MACHINE corresponds to an support machine estimator
        (valid for single-task classification problems)
    * ALL combines all estimator choices (valid for all learning tasks)
    """

    LINEAR = "LINEAR"
    RANDOM_FOREST = "RANDOM_FOREST"
    GAUSSIAN_PROCESS = "GAUSSIAN_PROCESS"
    SUPPORT_VECTOR_MACHINE = "SUPPORT_VECTOR_MACHINE"
    ALL = "ALL"


class AutoMLPredictor(Resource["AutoMLPredictor"], PredictorNode):
    """A predictor interface that builds a single ML model.

    The model uses the set of inputs to predict the output(s).
    Only one machine learning model is built.

    Parameters
    ----------
    name: str
        name of the configuration
    description: str
        the description of the predictor
    inputs: list[Descriptor]
        Descriptors that represent inputs to the model
    outputs: list[Descriptor]
        Descriptors that represents the output(s) of the model.
        Currently, only one output Descriptor is supported.
    estimators: Optional[Set[AutoMLEstimator]]
        Set of estimators to consider during during AutoML model selection.
        If None is provided, defaults to AutoMLEstimator.RANDOM_FOREST.
    training_data: Optional[List[DataSource]]
        Sources of training data. Each can be either a CSV or an GEM Table. Candidates from
        multiple data sources will be combined into a flattened list and de-duplicated by uid and
        identifiers. De-duplication is performed if a uid or identifier is shared between two or
        more rows. The content of a de-duplicated row will contain the union of data across all
        rows that share the same uid or at least 1 identifier. Training data is unnecessary if the
        predictor is part of a graph that includes all training data required by this predictor.

    """

    inputs = _properties.List(_properties.Object(Descriptor), 'inputs')
    outputs = _properties.List(_properties.Object(Descriptor), 'outputs')
    estimators = _properties.Set(
        _properties.Enumeration(AutoMLEstimator),
        'estimators',
        default={AutoMLEstimator.RANDOM_FOREST}
    )
    training_data = _properties.List(
        _properties.Object(DataSource),
        'training_data',
        default=[]
    )

    typ = _properties.String('type', default='AutoML', deserializable=False)

    def __init__(self,
                 name: str,
                 *,
                 description: str,
                 outputs: List[Descriptor],
                 inputs: List[Descriptor],
                 estimators: Optional[Set[AutoMLEstimator]] = None,
                 training_data: Optional[List[DataSource]] = None):
        self.name: str = name
        self.description: str = description
        self.inputs: List[Descriptor] = inputs
        self.estimators: Set[AutoMLEstimator] = estimators or {AutoMLEstimator.RANDOM_FOREST}
        self.outputs = outputs

        _check_deprecated_training_data(training_data)
        self.training_data: List[DataSource] = training_data or []

    def __str__(self):
        return '<AutoMLPredictor {!r}>'.format(self.name)



from typing import TypeVar, Optional, Union
from uuid import UUID

from citrine._serialization.serializable import Serializable
from citrine._serialization import properties
from gemd.entity.dict_serializable import DictSerializable
from gemd.enumeration.base_enumeration import BaseEnumeration
from gemd.util import make_index, substitute_objects


class ResourceTypeEnum(BaseEnumeration):
    """The type of the resource; used for modifying access controls.

    * TEAM is a Team
    * PROJECT is a Project
    * DATASET is a Dataset
    * MODULE is a Module: a Predictor or Design Space
    * USER is a user
    * TABLE is a GemTable
    * TABLE_DEFINITION is a TableConfig

    """

    TEAM = "TEAM"
    PROJECT = "PROJECT"
    DATASET = "DATASET"
    MODULE = "MODULE"
    USER = "USER"
    TABLE = "TABLE"
    TABLE_DEFINITION = "TABLE_DEFINITION"


Self = TypeVar('Self', bound='Resource')


class Resource(Serializable[Self]):
    """Abstract class for representing individual REST resource."""

    _path_template: str = NotImplemented
    _response_key: Optional[str] = None
    _resource_type: ResourceTypeEnum = NotImplemented

    def access_control_dict(self) -> dict:
        """Return an access control entity representation of this resource. Internal use only."""
        return {
            "type": self._resource_type.value,
            "id": str(self.uid)
        }


GEMDSelf = TypeVar('GEMDSelf', bound='GEMDResource')


class GEMDResource(Resource[GEMDSelf]):
    """A reference to a resource by UID."""

    @classmethod
    def build(cls, data: dict) -> GEMDSelf:
        """Convert a raw, nested dictionary into Objects."""
        if "context" in data and len(data) == 2:
            def _inflate(x):
                return DictSerializable.class_mapping[x["type"]].build(x)
            key = next(k for k in data if k != "context")
            idx = make_index([_inflate(x) for x in data["context"] + [data[key]]])
            lst = [idx[k] for k in idx]
            substitute_objects(lst, idx, inplace=True)

            root = _inflate(data[key])
            if root in idx:  # It was a link
                return idx[root]
            else:  # It was an object, but it won't be densely linked
                return idx[root.to_link()]
        else:
            if data.get("type") is not None:
                if not issubclass(cls, DictSerializable.class_mapping.get(data.get("type"))):
                    raise ValueError(f"{cls.__name__} passed a {data.get('type')} dictionary.")
            return super().build(data)

    def as_dict(self) -> dict:
        """
        Dump to a dictionary (useful for interoperability with gemd).

        Because of the _key mapping in Property, __dict__'s keys are fundamentally different
        between gemd.entity.dict_serializable and this class.  This means we can't just use gemd's
        as_dict for comparisons.
        """
        result = dict()
        for name, property_ in properties.Object(type(self)).fields.items():
            if property_.serializable:
                result[property_.serialization_path] = getattr(self, name, None)
        return result


class ResourceRef(Serializable['ResourceRef']):
    """A reference to a resource by UID."""

    # json key 'module_uid' is a legacy of when this object was only used for modules
    uid = properties.UUID('module_uid')

    def __init__(self, uid: Union[UUID, str]):
        self.uid = uid


class PredictorRef(Serializable['PredictorRef']):
    """A reference to a resource by UID."""

    uid = properties.UUID('predictor_id')
    version = properties.Optional(
        properties.Union([properties.Integer(), properties.String()]),
        'predictor_version'
    )

    def __init__(self, uid: Union[UUID, str], version: Optional[Union[int, str]] = None):
        self.uid = uid
        self.version = version
